# -*- coding: utf-8 -*-
#
# pygenda_util.py
# Miscellaneous utility functions for Pygenda.
#
# Copyright (C) 2022-2024 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.


from calendar import day_abbr,month_abbr
from icalendar import cal as iCal
from datetime import date, time, datetime, timedelta, tzinfo, timezone
from dateutil import tz as du_tz
from tzlocal import get_localzone
import locale
from typing import Tuple, Any, Union, Optional

from .pygenda_config import Config


def datetime_to_date(dt:date) -> date:
    # Extract date from datetime object (which might be a date)
    try:
        return dt.date()
    except AttributeError:
        return dt


def datetime_to_time(dt:date):
    # Extract time from datetime object.
    # Return False if no time component.
    try:
        return dt.time()
    except AttributeError:
        return False


# For some calculations, we need to have a timezone that is "aware" of
# summer time changes to make correct, e.g. hourly, calculations
try:
    _local_tz = get_localzone() # type:tzinfo
    if not hasattr(_local_tz,'unwrap_shim') and hasattr(_local_tz,'zone'):
        # Old tzlocal API, so need to create tzinfo object
        _local_tz = du_tz.gettz(_local_tz.zone) # type:ignore
except:
    # Fallback method for Linuxes:
    _local_tz = du_tz.tzfile('/etc/localtime')


def get_local_tz() -> Any:
    # Getter. Can be assumed to return a zoneinfo object.
    return _local_tz

def _set_local_tz(zinfo:Any) -> None:
    # Setter for test code - so we can run tests in different time zones.
    # Users should set time from the system, not an agenda app!
    global _local_tz
    _local_tz = zinfo


def date_to_datetime(dt:date, tz:Union[tzinfo,bool]=None) -> datetime:
    # Return datetime from dt argument. Set time to midnight if no time.
    # If tz parameter is not None/False, and dt has no timezone, add tz (True -> add local timezone)
    # Hence this function can be used to guarantee we have datetime with a timezone
    dt_ret= dt if isinstance(dt,datetime) else datetime(dt.year,dt.month,dt.day)
    if tz and dt_ret.tzinfo is None:
        dt_ret = dt_ret.replace(tzinfo=_local_tz if tz is True else tz)
    return dt_ret


def start_end_dts_event(event:iCal.Event) -> Tuple[date,date]:
    # Return start & end time of an event.
    # End time calculated from duration if needed; is None if no end/duration.
    start = event['DTSTART'].dt
    if 'DTEND' in event:
        end = event['DTEND'].dt
    elif 'DURATION' in event:
        end = start + event['DURATION'].dt
    else:
        end = None
    return start,end


def start_end_dts_occ(occ:Tuple[iCal.Event,date]) -> Tuple[date,date]:
    # Return start & end time of an occurrence (an (event,date[time]) pair)
    start = occ[1]
    if 'DTEND' in occ[0]:
        root_dt = occ[0]['DTSTART'].dt
        if isinstance(start, datetime):
            root_dt = date_to_datetime(root_dt,start.tzinfo)
        d = start - root_dt
        end = occ[0]['DTEND'].dt + d
    elif 'DURATION' in occ[0]:
        end = start + occ[0]['DURATION'].dt
    else:
        end = None
    return start,end


def format_time(dt, aslocal:bool=False) -> str:
    # Return time as string, formatted according to app config settings.
    # Argument 'dt' can be a time or a datetime.
    # If aslocal is True, convert time to local device time before formatting.
    if aslocal and dt.tzinfo:
        dt = dt.astimezone()
    if Config.get_bool('global','24hr'):
        fmt = '{hr:02d}{hs:s}{min:02d}'
        return fmt.format(hr=dt.hour, min=dt.minute, hs=Config.get('global','time_sep'))
    else:
        fmt = '{hr:d}{hs:s}{min:02d}{ampm:s}'
        return fmt.format(hr=(dt.hour-1)%12+1, min=dt.minute, ampm = 'pm' if dt.hour>11 else 'am', hs=Config.get('global','time_sep'))


def format_compact_time(dt, aslocal:bool=False) -> str:
    # Return time as string, according to app config settings.
    # Compact if 12hr set & minutes=0: e.g. '1pm' rather than '1:00pm'
    # Argument 'dt' can be a time or a datetime.
    # If aslocal is True, convert time to local device time before formatting.
    if aslocal and dt.tzinfo:
        dt = dt.astimezone()
    if Config.get_bool('global','24hr'):
        fmt = '{hr:02d}{hs:s}{min:02d}'
        return fmt.format(hr=dt.hour, min=dt.minute, hs=Config.get('global','time_sep'))
    else:
        if dt.minute:
            fmt = '{hr:d}{hs:s}{min:02d}{ampm:s}'
        else:
            fmt = '{hr:d}{ampm:s}'
        return fmt.format(hr=(dt.hour-1)%12+1, min=dt.minute, ampm = 'pm' if dt.hour>11 else 'am', hs=Config.get('global','time_sep'))


def format_compact_date(dt:date, show_year:bool, aslocal:bool=False) -> str:
    # Return date as string, with abbreviated day/month names.
    # If aslocal is True, convert time to local device time before formatting.
    if aslocal and isinstance(dt,datetime) and dt.tzinfo:
        dt = dt.astimezone()
    if show_year:
        fmt = '{day:s} {date:d} {mon:s}, {yr:d}'
    else:
        fmt = '{day:s} {date:d} {mon:s}'
    return fmt.format(day=day_abbr[dt.weekday()], date=dt.day, mon=month_abbr[dt.month], yr=dt.year)


def format_compact_datetime(dt:datetime, show_year:bool, aslocal:bool=False) -> str:
    # Return datetime as string, with abbreviated day/month names, compact time.
    # If aslocal is True, convert time to local device time before formatting.
    if aslocal and isinstance(dt,datetime) and dt.tzinfo:
        dt = dt.astimezone()
    if show_year:
        fmt = '{day:s} {date:d} {mon:s}, {yr:d}, {tm:s}'
    else:
        fmt = '{day:s} {date:d} {mon:s}, {tm:s}'
    return fmt.format(day=day_abbr[dt.weekday()], date=dt.day, mon=month_abbr[dt.month], yr=dt.year, tm=format_compact_time(dt))


def tzinfo_display_name(idt) -> Optional[str]:
    # Return name for timezone given idt, a datetime element of an iEvent/iTodo
    if 'TZID' in idt.params:
        nm = idt.params['TZID'] # type:str
        if nm:
            sl = nm.rfind('/')
            if sl>=0:
                nm = nm[sl+1:]
            nm = nm.replace('_',u' ')
            return nm
    if idt.dt.tzinfo is not None:
        return(str(idt.dt.tzinfo))
    return None


def day_in_week(dt:date) -> int:
    # Return the day number of dt in the week
    # Depends on config setting global/start_week_day
    day = dt.weekday()
    start_day = Config.get_int('global', 'start_week_day')
    return (day-start_day)%7


def start_of_week(dt:date) -> date:
    # Return first day of week containing dt
    # Depends on config setting global/start_week_day
    return dt - timedelta(days=day_in_week(dt))


def dt_lte(dt_a:date, dt_b:date) -> bool:
    # Return True if dt_a <= dt_b
    # dt_a and dt_b can be dates/datetimes independently.
    # The main purpose of this function is to sort events for display.
    # Hence if comparing a date to a datetime, take date as midnight
    # at start of day in local timezone.
    if isinstance(dt_a,datetime) or isinstance(dt_b,datetime):
        return date_to_datetime(dt_a,True) <= date_to_datetime(dt_b,True)
    return dt_a <= dt_b


def dt_lt(dt_a:date, dt_b:date) -> bool:
    # Return True if dt_a < dt_b
    # dt_a and dt_b can be dates/datetimes independently.
    # The main purpose of this function is to sort events for display.
    # Hence if comparing a date to a datetime, take date as midnight
    # at start of day in local timezone.
    if isinstance(dt_a,datetime) or isinstance(dt_b,datetime):
        return date_to_datetime(dt_a,True) < date_to_datetime(dt_b,True)
    return dt_a < dt_b


# We want to be able to sort events/todos by datetime.
# These functions return the relevant datetime for comparison.
def _event_sort_dt(self) -> date:
    return self['DTSTART'].dt # type:ignore[no-any-return]

def _todo_sort_dt(self) -> date:
    return self['DUE'].dt # type:ignore[no-any-return]

iCal.Event._sort_dt = _event_sort_dt
iCal.Todo._sort_dt = _todo_sort_dt


# And these functions do the comparison between events/todos.
def _event_lt(self, other) -> bool:
    # Return True if start date/time of self < start date/time of other
    return dt_lt(self['DTSTART'].dt, other._sort_dt())

def _todo_lt(self, other) -> bool:
    # Return True if start date/time of self < start date/time of other
    return dt_lt(self['DUE'].dt, other._sort_dt())

# Attach methods to classes so they will be used to sort
iCal.Event.__lt__ = _event_lt
iCal.Todo.__lt__ = _todo_lt


def dt_add_delta(dt:date, delta:timedelta) -> datetime:
    # Return dt+delta, taking into account things like daylight savings
    # changeover etc.
    r = date_to_datetime(dt, True)
    if delta.days:
        # Add days with timezone - at DST crossover days can have 23/25 hours
        r += timedelta(days=delta.days)
        delta = timedelta(seconds=delta.seconds,microseconds=delta.microseconds)
    # Add seconds in UTC because then there's no DST to worry about
    r_utc = r.astimezone(timezone.utc)
    r_utc += delta
    # Convert back to localtime for return
    r = r_utc.astimezone(_local_tz)
    return r


def guess_date_ord_from_locale() -> str:
    # Try to divine order of date (day/mon/yr mon/day/yr etc.) from locale.
    # Return a string like 'DMY' 'MDY' etc.
    dfmt = locale.nl_langinfo(locale.D_FMT) # E.g. '%x', '%m/%d/%Y'
    # To cope with dfmt=='%x': format a time & look at result
    st = date(year=3333,month=11,day=22).strftime(dfmt)
    ret = ''
    for ch in st:
        if ch=='3':
            if not ret or ret[-1]!='Y':
                ret += 'Y'
        elif ch=='1':
            if not ret or ret[-1]!='M':
                ret += 'M'
        elif ch=='2':
            if not ret or ret[-1]!='D':
                ret += 'D'
    return ret if len(ret)==3 else 'YMD'


def guess_date_sep_from_locale() -> str:
    # Try to divine date separator (/, -, etc.) from locale.
    dfmt = locale.nl_langinfo(locale.D_FMT)
    # To cope with dfmt=='%x': format a time & look at result
    st = date(year=1111,month=11,day=11).strftime(dfmt)
    for ch in st:
        if ch!='1':
            return ch
    return '-' # fallback


def guess_time_sep_from_locale() -> str:
    # Try to divine time separator (:, etc.) from locale.
    tfmt = locale.nl_langinfo(locale.T_FMT) # E.g. '%r', '%H:%M:%S'
    # To cope with tfmt=='%x': format a time & look at result
    st = time(hour=11,minute=11,second=11).strftime(tfmt)
    for ch in st:
        if ch!='1':
            return ch
    return ':' # fallback


def guess_date_fmt_text_from_locale() -> str:
    # Try to divine date formatting string from locale
    dtfmt = locale.nl_langinfo(locale.D_T_FMT) # includes time element
    ret = ''
    i = 0
    incl = True # including chars
    fmtch = False # formatting character
    while i < len(dtfmt):
        ch = dtfmt[i]
        if ch=='%' and not fmtch and dtfmt[i+1] in 'HIMRSTXZfklnprstz':
            incl = False # skip these until space char
        if incl:
            if fmtch and ch in 'aby':
                ch = ch.upper() # uppercase version for full day/month/year
            ret += ch
            fmtch = (ch=='%') # Boolean - next char is formatting char
        if ch==' ':
            incl = True
        i += 1
    return ret if ret else '%A, %B %-d, %Y' # fallback


def parse_timedelta(s:str) -> timedelta:
    # Parse a sting like "1h30m" to timedelta.
    # Used to parse config file entries.
    MAP = {
        'd': lambda a: timedelta(days=a),
        'h': lambda a: timedelta(hours=a),
        'm': lambda a: timedelta(minutes=a),
        's': lambda a: timedelta(seconds=a),
        }
    s = s.lower()
    td = timedelta()
    i = 0
    while i < len(s):
        j = i
        while s[j] not in MAP:
            j += 1
            if j == len(s):
                return td
        try:
            n = int(s[i:j])
        except ValueError:
            break
        td += MAP[s[j]](n)
        i = j+1
    return td


def utc_now_stamp() -> datetime:
    # Return current time suitable for use as an iCalendar timestamp.
    # Spec says it must be UTC. iCal stuctures go to second accuracy.
    return datetime.now(timezone.utc).replace(microsecond=0)



# -*- coding: utf-8 -*-
#
# pygenda_view.py
# "View" class definition - base class for Week/Year/Todo View.
# Provides default implementations of functions.
#
# Copyright (C) 2022-2024 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.


from gi.repository import Gtk, Gdk, GLib
from gi.repository.Pango import WrapMode as PWrapMode

from icalendar import cal as iCal, Event as iEvent, Todo as iTodo
from datetime import date as dt_date, datetime as dt_datetime, timedelta
from typing import Optional, Union

from .pygenda_gui import GUI
from .pygenda_config import Config
from .pygenda_dialog_event import EventDialogController
from .pygenda_util import datetime_to_time, datetime_to_date, date_to_datetime, format_time, format_compact_date, format_compact_time, format_compact_datetime, dt_lte, get_local_tz, tzinfo_display_name
from .pygenda_calendar import Calendar
from .pygenda_entryinfo import EntryInfo


# Base class for pygenda Views (Week View, Year View, etc.)
class View:
    # Provide shared variables & default implementations of required methods.

    # Variables for cursor display/navigation
    _cursor_date = dt_date.today()
    _cursor_idx_in_date = 0 # cursor index within date
    _today_toggle_date = None
    _today_toggle_idx = 0

    @staticmethod
    def view_name() -> str:
        # Return name string to use in menu, should be localised.
        # This default should never be called, so provide dummy name.
        return 'Null View'


    @staticmethod
    def accel_key() -> int:
        # Return keycode for menu shortcut, should be localised.
        return 0


    @classmethod
    def init(cls) -> Gtk.Widget:
        # Initialise view and return Gtk widget for that view.
        # Called from GUI._init_views() on startup.
        return None


    @classmethod
    def init_zoom(cls, config_gp:str, ctx:Gtk.StyleContext) -> None:
        # Initialise zoom level
        # Called during initialisation by child class.
        cls.zoom_lvls = Config.get_int(config_gp,'zoom_levels') # type:ignore
        cls.zoom_lvl = Config.get_int(config_gp,'default_zoom') # type:ignore
        cls.zoom_lvl %= cls.zoom_lvls # type:ignore
        cls.zoom_ctx = ctx # type:ignore
        cls.zoom_ctx.add_class('zoom'+str(cls.zoom_lvl)) # type:ignore


    @classmethod
    def renew_display(cls) -> None:
        # Called when we switch to this view.
        # Used to reset local state; default null implementation may be enough.
        pass


    @classmethod
    def redraw(cls, en_changes:bool) -> None:
        # Called when redraw required
        # en_changes: bool indicating if displayed entries need updating too
        pass


    @classmethod
    def keypress(cls, wid:Gtk.Widget, ev:Gdk.Event) -> None:
        # Called (from GUI.keypress()) on keypress (or repeat) event
        # Default does nothing. Derived views will override.
        pass


    @classmethod
    def keyrelease(cls, wid:Gtk.Widget, ev:Gdk.Event) -> None:
        # Called (from GUI.keypress()) on key release event
        # Default does nothing. Derived views may override.
        pass


    @classmethod
    def default_entry_is_todo(cls) -> bool:
        # Returns True if creating an "entry" in this view would
        # mean creating a Todo item.
        # Default is False - an "entry" would be an Event.
        return False


    @classmethod
    def paste_entry(cls, en:Union[iEvent,iTodo]) -> None:
        # Creates new entry based on entry en.
        # Type of entry should depend on View (e.g. Todo View -> to-do item).
        # Default implementation does nothing.
        pass


    @classmethod
    def paste_text(cls, txt:str) -> None:
        # Handle pasting of text.
        # Default implementation does nothing.
        pass


    @classmethod
    def cursor_inc(cls, delta:timedelta, idx:int=None) -> None:
        # Add delta to current cursor date; optionally set index in date.
        # Call redraw on view.
        View._cursor_date += delta # raises OverflowError if out of range
        if idx is not None:
            View._cursor_idx_in_date = idx
        cls.redraw(en_changes=False)


    @classmethod
    def cursor_set_date(cls, dt:dt_date) -> bool:
        # Set cursor date.
        # Return True if can (False if can't) jump to date in this view.
        # Default implementation does nothing & returns False.
        return False


    @classmethod
    def cursor_date(cls) -> Optional[dt_date]:
        # Returns date (maybe datetime in the future) with cursor.
        # Default implementation: cursor not on date.
        return None


    @classmethod
    def cursor_todo_list(cls) -> Optional[int]:
        # returns index of todo list with cursor
        # Default implementation: cursor not on todo list.
        return None


    @classmethod
    def cursor_goto_event(cls, ev:iEvent) -> bool:
        # Move cursor to given event.
        # Return True if can (False if can't) jump to event in this view.
        # Default implementation does nothing & returns False.
        return False


    @classmethod
    def cursor_goto_todo(cls, todo:iTodo, list_idx:int) -> bool:
        # Move cursor to given todo in given list.
        # Return True if can (False if can't) jump to todo in this view.
        # Default implementation does nothing & returns False.
        return False


    @classmethod
    def zoom(cls, inc:int) -> None:
        # Zoom by inc, so zoom-in if inc==+1; zoom-out if inc==-1
        # Assumes zoom_init() has been called.
        # If not using this system, View should supply its own zoom().
        # !! Note: might be better if it kept the cursor visible
        cls.zoom_ctx.remove_class('zoom'+str(cls.zoom_lvl)) # type:ignore
        cls.zoom_lvl = (cls.zoom_lvl+inc)%cls.zoom_lvls # type:ignore
        cls.zoom_ctx.add_class('zoom'+str(cls.zoom_lvl)) # type:ignore


    @staticmethod
    def entry_text_label(ev:iCal.Event, dt_st:dt_date, dt_end:dt_date, add_location:bool=False, loc_max_chars:int=0) -> Gtk.Label:
        # Returns a GtkLabel with entry summary + icons as content.
        # Used by Week & Year views to display entries.
        lab = Gtk.Label()
        lab.set_line_wrap(True)
        lab.set_line_wrap_mode(PWrapMode.WORD_CHAR)
        lab.set_xalign(0)
        lab.set_yalign(0)
        endtm = View.entry_endtime(dt_st,dt_end,True)
        icons = View.entry_icons(ev,True)
        d_txt = ev['SUMMARY'] if 'SUMMARY' in ev else ''
        if add_location and 'LOCATION' in ev:
            loc = ev['LOCATION']
            if loc_max_chars>0 and len(loc)>loc_max_chars:
                loc = loc[:loc_max_chars] + u'…'
            l_txt = ' (@{:s})'.format(loc)
        else:
            l_txt = ''
        z_txt = ''
        if isinstance(dt_st,dt_datetime) and dt_st.tzinfo is not None and dt_st.utcoffset()!=dt_st.astimezone(get_local_tz()).utcoffset():
            z_nm = tzinfo_display_name(ev['DTSTART'])
            if z_nm:
                z_tm = format_time(dt_st)
                z_txt = u'({:s} {:s}) '.format(z_tm,z_nm)
        lab.set_text(u'{:s}{:s}{:s}{:s}{:s}'.format(z_txt,d_txt,endtm,l_txt,icons))
        if isinstance(ev, iEvent):
            View.add_event_styles(lab, ev)
        elif isinstance(ev, iTodo):
            View.add_todo_styles(lab, ev)
        return lab


    ICON_NOTES = u'✉' # alternatives: ◻☐🀙⊟🗈🗎▤, enclosing square⃞ with ≡
    ICON_ALARM = u'♫' # alternatives: alarm clock ⏰ (U+23F0), bell 🕭,🔔 (U+1F56D,U+1F514), speaker 🔊
    ICON_REPEAT = u'⟳'

    @staticmethod
    def entry_icons(en:Union[iEvent,iTodo], prefix_space:bool) -> str:
        # Returns string of icons for entry (repeat, alarm...)
        icons = ''
        if 'DESCRIPTION' in en:
            icons += View.ICON_NOTES
        if en.walk('VALARM'):
            icons += View.ICON_ALARM
        if 'RRULE' in en:
            icons += View.ICON_REPEAT
        if prefix_space and icons:
            icons = u' {:s}'.format(icons) # thin space
        return icons


    @staticmethod
    def entry_endtime(dt_st:dt_date, dt_end:dt_date, frame_text:bool) -> str:
        # Returns end time in format suitable for displaying in entry.
        if not dt_end or dt_lte(dt_end, dt_st):
            return ''
        if type(dt_st) is dt_date and type(dt_end) is dt_date:
            end_date = dt_end-timedelta(1)
            if dt_st>=end_date:
                return ''
            t_str = format_compact_date(end_date, dt_st.year!=end_date.year, True) # type:str
        else:
            # At least one of dt_st,st_end is a datetime
            # => consider them *both* as datetimes
            st_date = datetime_to_date(dt_st)
            dt_end = date_to_datetime(dt_end)
            end_date = datetime_to_date(dt_end-timedelta(milliseconds=1))
            if st_date == end_date:
                t_str = format_compact_time(dt_end, True)
            else:
                t_str = format_compact_datetime(dt_end, st_date.year!=end_date.year, True)
        if frame_text and t_str:
            t_str = u' (→{:s})'.format(t_str)
        return t_str


    @staticmethod
    def add_event_styles(wid:Gtk.Label, ev:iCal.Event) -> None:
        # Adds formatting class to label corresponding to event status.
        # Allows entry to be formatted appropriately by CSS.
        ctx = wid.get_style_context()
        ctx.add_class('event')
        if 'STATUS' in ev and ev['STATUS'] in Calendar.STATUS_LIST_EVENT:
            ctx.add_class(ev['STATUS'].lower())


    @staticmethod
    def add_todo_styles(wid:Gtk.Label, td:iCal.Todo) -> None:
        # Adds formatting class to label corresponding to todo status.
        # Allows entry to be formatted appropriately by CSS.
        ctx = wid.get_style_context()
        ctx.add_class('todo')
        if 'STATUS' in td and td['STATUS'] in Calendar.STATUS_LIST_TODO:
            ctx.add_class(td['STATUS'].lower())


    @staticmethod
    def scroll_to_row(rowbox:Gtk.Box, row:int, scroller:Gtk.ScrolledWindow) -> bool:
        # Given a box displaying entries in a scroller (such as those
        # in Week & Year views), scroll to reveal row number 'row'.
        # Note: this may need to be called after rows are rendered,
        #   so that the allocated_heights are correct
        top = rowbox.get_spacing()*row # to hold max value to show top of cell
        rows = rowbox.get_children()
        for i in range(row):
            top += rows[i].get_allocated_height()
        bot = top + rows[row].get_allocated_height()
        bot -= scroller.get_allocated_height()
        # Account for padding, margin, border
        # First get the widths from the style context
        ctx = rowbox.get_style_context()
        pad = ctx.get_padding(Gtk.StateFlags.NORMAL)
        bord = ctx.get_border(Gtk.StateFlags.NORMAL)
        marg = ctx.get_margin(Gtk.StateFlags.NORMAL)
        # Looks a bit odd to be exactly on edge, so go halfway into padding
        top += bord.top + marg.top + (pad.top+1)//2
        bot += pad.top + bord.top + marg.top + (pad.bottom+1)//2
        if bot>top: # need this if row height bigger than scroll area
            bot = top # to favour showing top, rather than bottom, of cell
        adj = scroller.get_vadjustment()
        v = adj.get_value()
        if top < v:
            adj.set_value(top)
        elif v < bot:
            adj.set_value(bot)
        # According to:
        # https://developer.gnome.org/glib/stable/glib-The-Main-Event-Loop.html#g-idle-add
        # returning False is the right thing to do for one-shot calls.
        # Not sure if this is required, but let's do it anyway.
        return False


    @staticmethod
    def y_to_day_row(rowbox:Gtk.Box, y:float, maxrow:int, scroller:Gtk.ScrolledWindow=None) -> int:
        # Helper function to convert y-coord to row index in day.
        # Used to calculate entry to jump to when entry list is clicked,
        # for example in Week or Year View.
        if maxrow <= 1:
            return 0
        rs = rowbox.get_spacing()
        # Take account of padding/border/margins/row-spacing/scroller...
        ctx = rowbox.get_style_context()
        yc = ctx.get_padding(Gtk.StateFlags.NORMAL).top
        yc += ctx.get_border(Gtk.StateFlags.NORMAL).top
        yc += ctx.get_margin(Gtk.StateFlags.NORMAL).top
        yc -= rs/2
        if scroller is not None:
            yc -= scroller.get_vadjustment().get_value()

        row = -1 # return value
        rows = rowbox.get_children()
        # accumulate height of rows until it's greater than target y
        for i in range(maxrow):
            row += 1
            yc += rows[i].get_allocated_height()
            yc += rs
            if yc > y:
                break
        return row


    @staticmethod
    def remove_all_classes(ctx:Gtk.StyleContext) -> None:
        # Helper function to remove all classes from a view context.
        # Used when redrawing Year view, might be used for other views.
        for c in ctx.list_classes():
            ctx.remove_class(c)


class View_DayUnit_Base(View):
    # Used as base class for Day & Year views

    show_todos = False # Set to True by child class if shows todos
    _target_entry = None

    @classmethod
    def paste_entry(cls, en:Union[iEvent,iTodo]) -> None:
        # Creates new entry based on entry en.
        # Implementation for Week and Year Views - makes an event.
        new_en = Calendar.paste_entry(en, e_type=EntryInfo.TYPE_EVENT, dt_start=View._cursor_date)
        cls.cursor_goto_event(new_en)


    @classmethod
    def paste_text(cls, txt:str) -> None:
        # Handle pasting of text in Week/Year views.
        # Open a New Entry dialog with description initialised as txt
        date = cls.cursor_date()
        GLib.idle_add(EventDialogController.new_event, txt, date)


    @classmethod
    def cursor_date(cls) -> Optional[dt_date]:
        # Returns date(time) with cursor.
        # Default implementation: cursor not on date.
        return View._cursor_date


    @classmethod
    def cursor_edit_entry(cls) -> None:
        # Opens an entry edit dialog for the entry at the cursor,
        # or to create a new entry if the cursor is not on entry.
        # Assigned to the 'Enter' key in Week and Year views.
        en = cls.get_cursor_entry()
        if en is not None:
            if isinstance(en, iTodo):
                GUI.edit_or_display_todo(en)
            else:
                GUI.edit_or_display_event(en)
        elif GUI.create_events:
            date = cls.cursor_date()
            EventDialogController.new_event(date=date)


    @classmethod
    def get_cursor_entry(cls) -> Optional[iCal.Event]:
        # Returns entry at cursor position, or None if cursor not on entry.
        # Default: None. Derived classes will provided their implementations.
        return None


    @classmethod
    def cursor_goto_event(cls, ev:iEvent) -> bool:
        # Move cursor to given event.
        # Set target, so can jump on next redraw.
        repeats = 'RRULE' in ev # Boolean
        if not repeats:
            # Move cursor to first day of event
            # !! Need to make repeating case do something sensible
            dt = ev['DTSTART'].dt
            if isinstance(dt, dt_datetime):
                if dt.tzinfo is not None:
                    # dt has a timezone, so convert to local time
                    dt = dt.astimezone(get_local_tz())
                dt = dt.date()
            cls.cursor_set_date(dt)
        cls._target_entry = ev
        return True # Indicates success, so use this view


    @classmethod
    def cursor_goto_todo(cls, todo:iTodo, list_idx:int) -> bool:
        # Move cursor to given todo.
        # Sets target, so can jump on next redraw.
        # Returns True if moved to Todo entry, False to
        # indicate the caller should try another view.
        if not cls.show_todos:
            return False # Not showing any todos - try another view
        if 'DUE' not in todo:
            return False
        dt = todo['DUE'].dt
        if isinstance(dt, dt_datetime):
            if dt.tzinfo is not None:
                # dt has a timezone, so convert to local time
                dt = dt.astimezone(get_local_tz())
            dt = dt.date()
        cls.cursor_set_date(dt)
        cls._target_entry = todo
        return True # Indicates success, so use this view


    # Bullets to use as markers in Week/Year views
    _BULLET = u'•'
    _BULLET_ALLDAY = u'✦' # alternatives:❖⍟✪⦿❂♦⧫⏣⎔⌾⌘⌑⎊⎈⁕⧓⚫⚭⚙✷✦
    _BULLET_MULTIDAY_START = u'‣'
    _BULLET_ONGOING = u'»'
    _BULLET_TODO = u'Ⓣ' # alternative:🅣

    @classmethod
    def marker_label(cls, ev:iCal.Event, dt_st:dt_date, is_ongoing:bool=False) -> Gtk.Label:
        # Returns bullet or entry time suitable for marking entries.
        # Used to display entries in Week and Year views.
        lab = Gtk.Label()
        lab.set_halign(Gtk.Align.END)
        lab.set_valign(Gtk.Align.START)
        if is_ongoing:
            mark = cls._BULLET_ONGOING
        elif datetime_to_time(dt_st)!=False:
            mark = format_time(dt_st, True)
        elif type(ev) is iCal.Todo:
            mark = cls._BULLET_TODO
        elif 'DTEND' in ev:
            if dt_lte(ev['DTEND'].dt, ev['DTSTART'].dt+timedelta(days=1)):
                mark = cls._BULLET_ALLDAY
            else:
                mark = cls._BULLET_MULTIDAY_START
        else:
            mark = cls._BULLET
        lab.set_text(mark)

        return lab



# -*- coding: utf-8 -*-
#
# pygenda_dialog_entryprops.py
# Code for dialog to show entry properties
#
# Copyright (C) 2023,2024 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.


from gi.repository import Gtk
from gi.repository.Pango import WrapMode as PWrapMode

from datetime import datetime as dt_datetime, timedelta
from icalendar import Event as iEvent, Todo as iTodo, vDDDTypes
from locale import gettext as _
from typing import Union

# pygenda components
from .pygenda_gui import GUI
from .pygenda_calendar import Calendar


class EntryPropertiesDialog:
    # Dialog to display properties of 'entry' (an Event or Todo item)

    # Constants
    ENTRY_TYPES = ('Event','Todo','Journal')
    RRULE_RPTPROPLIST = {
        'BYMONTH': 'By month:',
        'BYWEEKNO': 'By week number:',
        'BYYEARDAY': 'By year day:',
        'BYMONTHDAY': 'By month day:',
        'BYDAY': 'By day:',
        'BYHOUR': 'By hour:',
        'BYMINUTE': 'By minute:',
        'BYSECOND': 'By second:',
        'BYSETPOS': 'By set position:',
        'WKST': 'Week starts:',
        'UNTIL': 'Until:',
        'COUNT': 'Count:'
        }
    STATUS_MAP = {
        'IN-PROCESS': 'In progress',
        'NEEDS-ACTION': 'Action required'
        }


    def __init__(self, entry:Union[iEvent,iTodo], parent:Gtk.Window=None):
        self.entry = entry
        self.dialog = Gtk.Dialog(title=_('Entry properties'), parent=parent,
            flags=Gtk.DialogFlags.MODAL|Gtk.DialogFlags.DESTROY_WITH_PARENT,
            buttons=(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE))
        self.dialog.set_resizable(False)
        self.grid = Gtk.Grid()
        self.grid_y = 0
        self.dialog.get_content_area().add(self.grid)


    def _write_dialog_content(self) -> None:
        # Fill the dialog with properties of self.entry
        self._add_property_row('SUMMARY')
        self._add_type_row()
        self._add_calendar_row()
        self._add_datetime_rows()
        self._add_rrule_row()
        self._add_property_row('PRIORITY')
        self._add_status_row()
        self._add_property_row('LOCATION')
        self._add_alarm_rows()
        self._add_description_rows()
        self._add_url_row()
        self._add_categories_row()


    def _add_row(self, label:str, cont:Union[str,Gtk.Widget], wrap:bool=False) -> None:
        # Add text to bottom of grid in the form "label cont".
        # If `cont` is a Widget, add it to grid; if a str convert to label.
        propnm_lab = Gtk.Label(label)
        propnm_lab.set_halign(Gtk.Align.END)
        propnm_lab.set_yalign(0)
        propnm_lab.get_style_context().add_class(GUI.STYLE_TXTLABEL)
        self.grid.attach(propnm_lab, 0,self.grid_y, 1,1)

        if isinstance(cont, Gtk.Widget):
            # Assume caller has set any styling etc.
            self.grid.attach(cont, 1,self.grid_y, 1,1)
        else:
            cont_lab = Gtk.Label(cont)
            cont_lab.set_halign(Gtk.Align.START)
            cont_lab.set_yalign(0)
            cont_lab.set_xalign(0)
            cont_lab.get_style_context().add_class(GUI.STYLE_TXTPROP)
            if wrap:
                cont_lab.set_line_wrap(True)
                cont_lab.set_line_wrap_mode(PWrapMode.WORD_CHAR)
                cont_lab.set_max_width_chars(50) # otherwise it fills screen
            self.grid.attach(cont_lab, 1,self.grid_y, 1,1)

        self.grid_y += 1


    def _add_property_row(self, propnm:str) -> None:
        # Add simple property in a single row to bottom of grid
        if propnm in self.entry:
            self._add_row(_(propnm.capitalize()+':'), self.entry[propnm])


    def _add_type_row(self) -> None:
        # Add entry type property to bottom of grid
        typestr = str(type(self.entry))
        afterdot = typestr.rfind('.')+1
        endquote = typestr.rfind("'")
        if afterdot>0 and endquote>afterdot:
            typestr = typestr[afterdot:endquote]
            if typestr in self.ENTRY_TYPES:
                self._add_row(_('Type:'), _(typestr))


    def _add_calendar_row(self) -> None:
        # Add Calendar (i.e. name of data store) property to bottom of grid
        dn = Calendar.calendar_displayname(self.entry)
        self._add_row(_('Calendar:'), dn)


    def _add_datetime_rows(self) -> None:
        # Add datetime start/end/duration properties to bottom of grid
        self._add_datetime_row_if_exists('DTSTART', _('Start date:'), _('Start date/time:'))
        self._add_datetime_row_if_exists('DTEND', _('End date:'), _('End date/time:'))
        if 'DURATION' in self.entry:
            dur = self.entry['DURATION'].dt
            self._add_row(_('Duration:'), str(dur))
        self._add_datetime_row_if_exists('DUE', _('Due date:'), _('Due date/time:'))
        self._add_datetime_row_if_exists('COMPLETED', _('Completed date (noncompliant):'), _('Completed date/time:'))


    def _add_datetime_row_if_exists(self, ical_lab:str, ui_date_lab:str, ui_dt_lab:str) -> None:
        # If ical_lab date/time property exists, add it to bottom of grid
        if ical_lab in self.entry:
            dt = self.entry[ical_lab]
            dt_dt = dt.dt
            if isinstance(dt_dt, dt_datetime):
                self._add_row(ui_dt_lab, str(dt_dt)+self._tz_str(dt))
            else:
                self._add_row(ui_date_lab, str(dt_dt))


    @staticmethod
    def _tz_str(dt_st:vDDDTypes) -> str:
        # Returns string indicating timezone, to be appended to times
        params = dt_st.params
        if 'TZID' in params:
            return _(' (timezone: {:s})').format(params['TZID']) # type:ignore
        return ''


    def _add_rrule_row(self) -> None:
        # Add rrule properties to bottom of grid
        rep_info = ''
        if 'RRULE' in self.entry:
            rr = self.entry['RRULE']
            if 'FREQ' in rr:
                rep_info = rr['FREQ'][0].capitalize()
                if 'INTERVAL' in rr:
                    rep_info += ' ({} {})'.format(_('interval:'), rr['INTERVAL'][0])
            for by in self.RRULE_RPTPROPLIST:
                if by in rr:
                    if rep_info:
                        rep_info +='\n'
                    val = rr[by]
                    if len(val) == 1:
                        val = val[0]
                    rep_info += '{} {}'.format(_(self.RRULE_RPTPROPLIST[by]), str(val))
        if 'EXDATE' in self.entry:
            exdts = Calendar.caldatetime_tree_to_dt_list(self.entry['EXDATE'])
            if rep_info:
                rep_info +='\n'
            rep_info += _('Exception dates:')
            exdt_str = ''
            for dt in exdts:
                if exdt_str:
                    exdt_str += _(';')
                exdt_str += ' ' + str(dt)
            rep_info += exdt_str

        # Finally, display the info string
        if rep_info:
            self._add_row(_('Repeats:'), rep_info)


    def _add_status_row(self) -> None:
        # Add status property to bottom of grid
        if 'STATUS' in self.entry:
            stat = self.entry['STATUS']
            if stat in self.STATUS_MAP:
                stat = self.STATUS_MAP[stat]
            else:
                stat = stat.capitalize()
            self._add_row(_('Status:'), _(stat))


    def _add_alarm_rows(self) -> None:
        # Add alarm properties to bottom of grid
        alarms = self.entry.walk('VALARM')
        if alarms:
            a_info = ''
            for a in alarms:
                if a_info:
                    a_info += '\n'
                if 'TRIGGER' in a:
                    trig = a['TRIGGER'].dt
                    if isinstance(trig,timedelta) and trig<timedelta(0):
                        # More intuitive way to show negative deltas
                        a_info += u'−'
                        trig = -trig
                    a_info += str(trig)
                    if 'ACTION' in a:
                        act = a['ACTION'].capitalize()
                        a_info += ' ' + _(act)
                        if act=='Email':
                            if 'ATTENDEE' in a:
                                # May be more than one attendee
                                a_info += '\n('
                                attee = a['ATTENDEE']
                                if isinstance(attee, list):
                                    a_info += ', '.join(attee)
                                else:
                                    a_info += attee
                                a_info += ')'
                        if act in ('Display','Email'):
                            if 'DESCRIPTION' in a:
                                # Spec says only one description
                                a_info += '\n'
                                a_info += _(u'“{:s}”').format(a['DESCRIPTION'])
                        if act=='Audio' and 'ATTACH' in a:
                            a_info += '\n'
                            a_info += a['ATTACH']
                else:
                    # No Trigger (so breaks specs)
                    a_info += 'Unspecified'
            self._add_row(_('Alarms:'), a_info)


    def _add_description_rows(self) -> None:
        # Add description(s) to bottom of grid.
        # (I notice that in RFC 5545, descriptions must not occur more
        # than once for events/todos. However, I wrote code to handle
        # more than one, so I'm leaving it that way.)
        if 'DESCRIPTION' in self.entry:
            edesc = self.entry['DESCRIPTION']
            if isinstance(edesc, list):
                for d in edesc:
                    if 'LANGUAGE' in d.params:
                        self._add_row(_('Description ({}):').format(d.params['LANGUAGE']), d, wrap=True)
                    else:
                        self._add_row(_('Description:'), d, wrap=True)
            else:
                # !! Not sure how to access LANGUAGE parameter for single desc
                self._add_row(_('Description:'), edesc, wrap=True)


    def _add_url_row(self) -> None:
        # Add URL property to bottom of grid.
        # Use a Gtk.LinkButton so the URL is clickable.
        # Maybe security implications here if people import entries.
        if 'URL' in self.entry:
            url = self.entry['URL']
            if url:
                link = Gtk.LinkButton.new(url)
                link.set_halign(Gtk.Align.START)
                link.get_style_context().add_class(GUI.STYLE_TXTPROP)
                self._add_row(_('URL:'), link)


    def _add_categories_row(self) -> None:
        # Add categories property to bottom of grid.
        if 'CATEGORIES' in self.entry:
            catgs = self.entry['CATEGORIES']
            # If entry has multiple CATEGORIES lines, catgs will be a list
            if isinstance(catgs, list):
                ctmp = [ ', '.join(c.cats) for c in catgs ]
            else:
                ctmp = catgs.cats
            cstr = ', '.join(ctmp)
            self._add_row(_('Categories:'), cstr)


    def run(self) -> Gtk.ResponseType:
        # run the dialog
        self._write_dialog_content()
        self.dialog.show_all()
        return self.dialog.run()


    def destroy(self) -> None:
        # destroy the dialog
        self.dialog.destroy()



# -*- coding: utf-8 -*-
#
# pygenda_dialog_todo.py
# Code for Todo dialog (used to create/update to-do items)
#
# Copyright (C) 2022-2024 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.


from gi.repository import Gtk, Gdk

from icalendar import Todo as iTodo
from locale import gettext as _
from datetime import datetime as dt_datetime
from typing import Optional, Tuple, Callable

# pygenda components
from .pygenda_gui import GUI
from .pygenda_calendar import Calendar
from .pygenda_entryinfo import EntryInfo
from .pygenda_widgets import WidgetDate


# Exception used to indicate dialog can't display todo properties.
# In these cases there is a danger of the user accidentally changing
# the todo property.
class TodoPropertyBeyondEditDialog(Exception):
    pass


# Singleton class to manage Todo dialog
# !! There's overlap between this class and the EventDialogController class.
# !! Consider merging the shared elements.
class TodoDialogController:
    dialog = None # type: Gtk.Dialog
    wid_desc = None # type: Gtk.Entry
    wid_todolist = None # type: Gtk.ComboBoxText
    wid_calendar = None # type: Gtk.ComboBox
    wid_priority = None # type: Gtk.ComboBoxText
    wid_status = None # type: Gtk.ComboBoxText
    wid_tabs = None # type: Gtk.Notebook

    wid_duedate_switch = None # type: Gtk.Switch
    revealer_duedate = None # type:Gtk.Revealer
    wid_duedate = None # type:WidgetDate

    buf_notes = None # type: Gtk.TextBuffer
    buf_notes_scroller = None # type: Gtk.ScrolledWindow

    list_default_cats = None # type: list
    todolist_index = None # type: Callable

    @classmethod
    def init(cls) -> None:
        # Initialiser for TodoDialogController singleton class.
        # Called from GUI init_stage2().

        # Load UI file
        GUI.load_ui_file('dialog_todo.ui')

        # Connect signal handlers
        HANDLERS = {
            'notes_focusin': cls._notes_focus,
            'notes_focusout': cls._notes_focusloss,
            'notes_keypress': cls._notes_keypress,
            }
        GUI._builder.connect_signals(HANDLERS)

        # Get some references to dialog elements from UI file
        cls.dialog = GUI._builder.get_object('dialog_todo')
        if (not cls.dialog): # Sanity check
            raise NameError('Dialog Todo not found')

        cls.wid_desc = GUI._builder.get_object('entry_dialogtodo_desc')
        cls.wid_desc.connect('changed', cls._validated_field_changed)
        cls._init_todolists()
        cls.wid_priority = GUI._builder.get_object('combo_todo_priority')
        cls.wid_status = GUI._builder.get_object('combo_todo_status')
        cls.wid_tabs = GUI._builder.get_object('dialogtodo_tabs')
        cls._init_calendarfields()
        cls._init_duedate()
        cls._init_notes()

        # Delayed import to set function to get a todo's todo list
        from .pygenda_view_todo import View_Todo
        cls.todolist_index = View_Todo.first_list_index


    @classmethod
    def _init_todolists(cls) -> None:
        # Get reference to todo-list combobox & initialise its entries
        cls.wid_todolist = GUI._builder.get_object('combo_todo_list')
        todo_titles, cls.list_default_cats = GUI.todo_titles_default_cats()
        for t in todo_titles:
            cls.wid_todolist.append_text(t)


    @classmethod
    def _init_calendarfields(cls) -> None:
        # Initialise calendar combo-box.
        # Called on app startup.
        cls.wid_calendar = GUI._builder.get_object('combo_todo_calendar')
        cal_list = Calendar.calendar_displaynames_todo_rw()
        for id,dn in cal_list:
            cls.wid_calendar.append(str(id), dn)
        if len(cal_list) <= 1:
            # Hide option if there's only one choice
            cls.wid_calendar.hide()
            GUI._builder.get_object('label_todo_calendar').hide()


    @classmethod
    def _init_duedate(cls) -> None:
        # Get references to due date element and set up revealer
        cls.wid_duedate_switch = GUI._builder.get_object('switch_todo_duedate')
        cls.revealer_duedate = GUI._builder.get_object('revealer_duedate')
        cls.wid_duedate = WidgetDate()
        cls.revealer_duedate.add(cls.wid_duedate)
        cls.wid_duedate.show_all()
        cls.wid_duedate.connect('changed', cls._validated_field_changed)
        cls.wid_duedate_switch.connect('state-set', cls._toggle_duedate)
        cls.wid_duedate.get_style_context().add_class('hidden') # Remove -ve width warnings
        # Allow keyboard navigation from date widget to switch
        cls.wid_duedate.set_left_widget(cls.wid_duedate_switch)


    @classmethod
    def _init_notes(cls) -> None:
        # Get references to notes widget.
        # Called on app startup.
        wid_notes = GUI._builder.get_object('textview_dialogtodo_notes')
        cls.buf_notes = wid_notes.get_buffer()
        cls.buf_notes_scroller = wid_notes.get_parent()


    @classmethod
    def _toggle_duedate(cls, wid:Gtk.ComboBox, val:bool) -> bool:
        # Callback for "Due date" switch. Shows or hides date field.
        cls.revealer_duedate.set_reveal_child(val)
        ctx = cls.wid_duedate.get_style_context()
        if val:
            ctx.remove_class('hidden')
        else:
            ctx.add_class('hidden')
        return False # must propagate event for active CSS selector to apply


    @classmethod
    def new_todo(cls, txt:str=None, list_idx:int=None) -> None:
        # Called to implement "new todo" from GUI, e.g. menu
        cls.dialog.set_title(_('New To-do'))
        response,ei,list_idx = cls._do_todo_dialog(txt=txt, list_idx=list_idx)
        if response==Gtk.ResponseType.OK and ei.desc:
            td = Calendar.new_entry(ei)
            GUI.cursor_goto_todo(td, list_idx)
            GUI.view_redraw(en_changes=True)


    @classmethod
    def edit_todo(cls, todo:iTodo, list_idx:int=None) -> None:
        # Called to implement "edit todo" from GUI
        cls.dialog.set_title(_('Edit To-do'))
        response,ei,list_idx = cls._do_todo_dialog(todo=todo, list_idx=list_idx)
        if response==Gtk.ResponseType.OK:
            if ei.desc:
                Calendar.update_entry(todo, ei)
                GUI.cursor_goto_todo(todo, list_idx)
                GUI.view_redraw(en_changes=True)
            else: # Description text has been deleted in dialog
                GUI.dialog_deleteentry(todo)


    @classmethod
    def _do_todo_dialog(cls, todo:iTodo=None, txt:str=None, list_idx:Optional[int]=0) -> Tuple[int,EntryInfo,int]:
        # Do the core work displaying todo dialog and extracting result.
        cls._seed_fields(todo=todo, txt=txt, list_idx=list_idx)
        cls.wid_tabs.set_current_page(0) # first tab
        cls._reset_err_style() # clear error highlights
        discard_on_empty = (todo is None) # True if creating a new todo entry
        try:
            while True:
                response = cls.dialog.run()
                if response!=Gtk.ResponseType.OK:
                    break
                if discard_on_empty and cls._fields_are_defaults():
                    # New todo & nothing entered, so discard as a null entry
                    response = Gtk.ResponseType.CANCEL
                    break
                if cls._is_valid_todo(set_style=True):
                    break
                discard_on_empty = False # only allow discard first time
        finally:
            cls.dialog.hide()
        return response, cls._get_entryinfo(), cls.wid_todolist.get_active()


    @classmethod
    def _seed_fields(cls, todo:Optional[iTodo], txt:Optional[str], list_idx:Optional[int]=0) -> None:
        # Initialise fields when dialog is opened.
        # Data optionally from an existing todo, or text used as summary.

        # First, set to default values (possibly overwritten later)
        cls._set_fields_to_defaults()

        if todo is not None:
            cls._set_fields_from_todo(todo)
        elif txt is not None:
            cls.wid_desc.set_text(txt)
            cls.wid_desc.set_position(len(txt))

        if list_idx is not None:
            cls.wid_todolist.set_active(list_idx)
        elif todo is not None:
            idx = cls.todolist_index(todo)
            if idx is not None:
                cls.wid_todolist.set_active(idx)


    @classmethod
    def _set_fields_to_defaults(cls) -> None:
        # Set dialog fields to default values.
        # If you change this, also change _fields_are_defaults()
        # and maybe _is_valid_todo() & _reset_err_style() below.
        cls.wid_desc.set_text('')
        cls.wid_desc.grab_focus_without_selecting()
        cls.wid_todolist.set_active(0)
        cls.wid_calendar.set_active(0)
        cls.wid_priority.set_active(0)
        cls.wid_status.set_active(0)
        cls.wid_duedate_switch.set_active(False)
        cls.wid_duedate.set_date(GUI.cursor_date())
        cls.buf_notes.set_text('')


    @classmethod
    def _fields_are_defaults(cls) -> bool:
        # Return True if all fields are defaults, False if any non-default
        # Ignores To-do List field. No particular justification for this,
        # but if someone just changes that, then it's not really the
        # to-do item that they changed, just where they want to store it.
        if cls.wid_desc.get_text():
            return False
        if cls.wid_priority.get_active()!=0:
            return False
        if cls.wid_duedate_switch.get_active():
            return False
        if cls.wid_status.get_active()!=0:
            return False
        return True


    @classmethod
    def _set_fields_from_todo(cls, todo:iTodo) -> None:
        # Set dialog fields from a given todo item.
        # Assumes values have already been set to default values.
        if 'SUMMARY' in todo:
            cls.wid_desc.set_text(todo['SUMMARY'])
        cls.wid_desc.grab_focus() # also selects contents
        cls.wid_calendar.set_active_id(str(todo._cal_idx))
        if 'PRIORITY' in todo:
            p = int(todo['PRIORITY'])
            if 1<=p<=9:
                cls.wid_priority.set_active(p)
        if 'DUE' in todo:
            due = todo['DUE'].dt
            if isinstance(due,dt_datetime):
                raise TodoPropertyBeyondEditDialog('Editing todo with date+time DUE not supported')
            cls.wid_duedate_switch.set_active(True)
            cls.wid_duedate.set_date(due)
        if 'DTSTART' in todo or 'DTEND' in todo or 'COMPLETED' in todo:
            raise TodoPropertyBeyondEditDialog('Editing todo with unsupported date property')
        if 'STATUS' in todo:
            s = todo['STATUS']
            if s in Calendar.STATUS_LIST_TODO:
                cls.wid_status.set_active_id(s)
        if 'DESCRIPTION' in todo:
            cls.buf_notes.set_text(todo['DESCRIPTION'])
            # Move cursor & scollbar to start:
            cls.buf_notes.place_cursor(cls.buf_notes.get_start_iter())
            cls.buf_notes_scroller.get_vadjustment().set_value(0)


    @classmethod
    def _is_valid_todo(cls, set_style:bool=False) -> bool:
        # Function checks if todo dialog fields are valid.
        # Used when "OK" clicked in the dialog.
        # Returns True if all fields valid; False if *any* field is invalid.
        # If a field is invalid and set_style==True it also adds a
        # CSS class to the widget, so the error is visibly indicated.
        # If the event is valid then it removes the error style, so
        # the error indication will disappear (regardless of set_style).
        # Note: If you change this, also update _reset_err_style() below.

        ret = True

        # Check description is non-empty
        ctx = cls.wid_desc.get_style_context()
        if not cls.wid_desc.get_text():
            ret = False
            if set_style:
                ctx.add_class(GUI.STYLE_ERR)
        else:
            ctx.remove_class(GUI.STYLE_ERR)

        # If due date is on, check given date is valid
        ctx = cls.wid_duedate.get_style_context()
        if cls.wid_duedate_switch.get_active() and not cls.wid_duedate.is_valid_date():
            ret = False
            if set_style:
                ctx.add_class(GUI.STYLE_ERR)
        else:
            ctx.remove_class(GUI.STYLE_ERR)

        return ret


    @classmethod
    def _reset_err_style(cls) -> None:
        # Remove 'dialog_error' style class from todo dialog widgets where
        # it might be set, so any visible indications of errors are removed.
        # Used, for example, when the dialog is reinitialised.
        for w in (cls.wid_desc, cls.wid_duedate):
            w.get_style_context().remove_class(GUI.STYLE_ERR)


    @classmethod
    def _validated_field_changed(cls, wid:Gtk.Entry) -> bool:
        # Callback. Called when a "validated field" changes.
        # A "validated field" is one that is checked with _is_valid_todo().
        # Function removes any error state styling that is no longer needed.
        # It does *not* add any new error styling.
        cls._is_valid_todo(set_style=False)
        return False # propagate event


    @classmethod
    def _get_entryinfo(cls) -> EntryInfo:
        # Decipher entry fields and return info as an EntryInfo object.
        desc = cls.wid_desc.get_text()
        stat = cls.wid_status.get_active_id()
        cal_idx = int(cls.wid_calendar.get_active_id())
        ei = EntryInfo(type=EntryInfo.TYPE_TODO, cal_idx=cal_idx, desc=desc, status=stat)
        list_idx = cls.wid_todolist.get_active()
        ei.set_categories(cls.list_default_cats[list_idx])
        if cls.wid_duedate_switch.get_active():
            ei.set_duedate(cls.wid_duedate.get_date_or_none())
        ei.set_priority(cls.wid_priority.get_active())
        ldesc = cls.buf_notes.get_text(cls.buf_notes.get_start_iter(), cls.buf_notes.get_end_iter(), False)
        if ldesc:
            ei.set_longdesc(ldesc)
        return ei


    @classmethod
    def _notes_focus(cls, wid:Gtk.Widget, ev:Gdk.EventFocus) -> bool:
        # Handler for notes fields getting focus.
        # Set style (on parent element to include scrollbar)
        cls.buf_notes_scroller.get_style_context().add_class('focus')
        return False # propagate event


    @classmethod
    def _notes_focusloss(cls, wid:Gtk.Widget, ev:Gdk.EventFocus) -> bool:
        # Handler for notes field losing focus
        cls.buf_notes_scroller.get_style_context().remove_class('focus')
        return False # propagate event


    @classmethod
    def _notes_keypress(cls, wid:Gtk.Widget, ev:Gdk.EventKey) -> bool:
        # Handler for key-press/repeat when notes field is focused
        if ev.keyval == Gdk.KEY_Up:
            # If cursor at start of text, need to handle navigation
            cur = cls.buf_notes.get_property('cursor-position')
            if cur==0:
                cls.wid_tabs.grab_focus()
                return True # Event handled, don't propagate
        elif ev.keyval==Gdk.KEY_Return and ev.state&(Gdk.ModifierType.SHIFT_MASK|Gdk.ModifierType.CONTROL_MASK)==0: # not shift or control
            # Manually trigger default event on dialog box
            cls.dialog.response(Gtk.ResponseType.OK)
            return True # Don't propagate event

        return False # Propagate event



# -*- coding: utf-8 -*-
#
# pygenda_dialog_import.py
# Code for import dialogs
#
# Copyright (C) 2024 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.


from gi.repository import Gtk

from icalendar import Calendar as iCalendar, Event as iEvent, Todo as iTodo
from locale import gettext as _ # type:ignore[attr-defined]
from sys import stderr
from typing import Optional, Union, Tuple

# Pygenda components
from .pygenda_gui import GUI
from .pygenda_calendar import Calendar


# Singleton class to manage importing file
class ImportController:

    _dialog_grid = None # type:Gtk.Grid
    _dialog_y = 0

    @classmethod
    def import_flow(cls) -> None:
        # Called to import a file
        filenm = cls._get_file()
        if filenm:
            cls._import_file(filenm)


    @classmethod
    def _get_file(cls) -> Optional[str]:
        # Opens file chooser, returns filename or None
        chooser = Gtk.FileChooserNative.new(_('Import File'), GUI._window, Gtk.FileChooserAction.OPEN, _('_Import'), None)

        # Add filters
        filt_ical = Gtk.FileFilter()
        filt_ical.set_name(_('iCalendar files'))
        filt_ical.add_mime_type('text/calendar')
        chooser.add_filter(filt_ical)
        filt_none = Gtk.FileFilter()
        filt_none.set_name(_('All files'))
        filt_none.add_pattern('*')
        chooser.add_filter(filt_none)

        # Run dialog and return value depending on result
        res = chooser.run()
        if res==Gtk.ResponseType.ACCEPT:
            return chooser.get_filename() # type:ignore[no-any-return]
        return None


    @classmethod
    def _import_file(cls, filename:str) -> None:
        # Opens file as iCal file and imports each entry
        with open(filename, 'rb') as file:
            try:
                cal = iCalendar.from_ical(file.read())
            except:
                print('Error: Failed to parse ical file '+filename, file=stderr)
                return
        if cal.is_broken:
            print('Error: Non-conformant ical data '+filename, file=stderr)
            return

        entries = cal.walk('VEVENT')
        entries.extend(cal.walk('VTODO'))

        if not entries:
            print('Notice: No event or todo entries found in '+filename, file=stderr)
            return

        jump_to_en = None # Store the first entry imported - to move cursor to
        for en in entries:
            new_en = cls._process_entry(en)
            if new_en is False:
                # False indicates import has been cancelled, so stop here
                break
            if jump_to_en is None and new_en is not None:
                jump_to_en = new_en

        if jump_to_en is not None:
            if isinstance(jump_to_en, iEvent):
                GUI.cursor_goto_event(jump_to_en)
            else:  # is a Todo
                GUI.cursor_goto_todo(jump_to_en, 0)# 0 = find first occ of todo
            GUI.view_redraw(en_changes=True)


    @classmethod
    def _get_calendar_combobox(cls, is_event:bool) -> Union[bool,Gtk.ComboBox]:
        # Return a combobox for selecting calendar for events.
        # Returns True => Only one calendar available, so no combobox required.
        # Returns False => No calendars available, so can't import.
        if is_event:
            cal_list = Calendar.calendar_displaynames_event_rw()
        else: # is a todo
            cal_list = Calendar.calendar_displaynames_todo_rw()
        if len(cal_list) == 0:
            # We can't import this entry!
            return False
        if len(cal_list) == 1:
            # No need for a combobox when there's no choice to make
            return True
        combobox = Gtk.ComboBoxText()
        for id,dn in cal_list:
            combobox.append(str(id), dn)
        combobox.connect('key-press-event', GUI._combobox_keypress, Gtk.ResponseType.ACCEPT)
        return combobox


    @classmethod
    def _process_entry(cls, en:Union[iEvent,iTodo]) -> Union[iEvent,iTodo,bool,None]:
        # Import or skip entry in iCal file, depending on user interaction.
        # Return entry if entry imported, None if skipped, False if cancelled.
        res,cal = cls._import_entry_dialog(en)
        if res==Gtk.ResponseType.DELETE_EVENT:
            return False
        if res==Gtk.ResponseType.ACCEPT:
            new_en = Calendar.import_entry(en, cal_idx=cal)
            return new_en
        return None


    @classmethod
    def _import_entry_dialog(cls, en:Union[iEvent,iTodo]) -> Tuple[int,Optional[int]]:
        # Show/manage dialog to query importing single entry.
        # This will be shown for each entry in an imported iCal file.
        dialog = Gtk.Dialog(title=_('Import entry'), parent=GUI._window,
            flags=Gtk.DialogFlags.MODAL|Gtk.DialogFlags.DESTROY_WITH_PARENT,
            buttons=(_('Skip'), Gtk.ResponseType.REJECT))
        dialog.set_resizable(False)
        import_button = dialog.add_button(_('Import'), Gtk.ResponseType.ACCEPT)
        can_import = True # Will use this to enable/disable Import button

        cls._dialog_grid = Gtk.Grid()
        dialog.get_content_area().add(cls._dialog_grid)
        cls._dialog_y = 0

        cls._add_row(_('Found entry:'), style=GUI.STYLE_SECTLABEL, halign=Gtk.Align.START)

        if 'SUMMARY' in en:
            desc_txt = en['SUMMARY']
            if len(desc_txt)>60:
                desc_txt = desc_txt[:60]+u'…'
        else:
            desc_txt = _('None')
        cls._add_row_prop(_('Description:'), desc_txt)
        en_is_event = isinstance(en,iEvent) # False => a Todo
        cls._add_row_prop(_('Type:'), _('Event') if en_is_event else _('Todo'))

        # Display extra information rows depending on entry type
        if en_is_event:
            if 'DTSTART' in en:
                cls._add_row_prop(_('Date/time:'), str(en['DTSTART'].dt))
            else:
                cls._add_row(_('Invalid event, no date found'))
                can_import = False
            if can_import and 'RRULE' in en:
                rr = en['RRULE']
                if 'FREQ' in rr:
                    rep_type = rr['FREQ'][0].capitalize()
                    cls._add_row_prop(_('Repeats:'), _(rep_type))
        else: # en is Todo
            if 'DUE' in en:
                cls._add_row_prop(_('Due date:'), str(en['DUE'].dt))

        # Check entry UID. If present, use it to test if entry already exists
        if 'UID' not in en:
            cls._add_row(_('Invalid entry, no ID found'), style=GUI.STYLE_ALERTLABEL, halign=Gtk.Align.CENTER)
            can_import = False
        elif Calendar.get_entry_by_uid(en['UID']) is not None: # entry exists?
            cls._add_row(_('An entry with this ID already exists'), style=GUI.STYLE_ALERTLABEL, halign=Gtk.Align.CENTER)
            can_import = False

        # If necessary, add dropdown box to select calendar to import into
        cb_cal = False
        if can_import:
            cb_cal = cls._get_calendar_combobox(en_is_event)
            if cb_cal is False:
                cls._add_row(_('No calendar available to store {:s}').format(_('an event') if en_is_event else _('a todo')), style=GUI.STYLE_ALERTLABEL, halign=Gtk.Align.CENTER)
                can_import = False
            elif cb_cal is not True:
                cb_cal.set_active(0) # type:ignore[union-attr]
                cls._add_row_widget(_('Import to calendar:'), cb_cal)

        # If can_import, sensitise the "Import" button
        import_button.set_sensitive(can_import)

        dialog.show_all()
        res = dialog.run() # type:int
        dest_cal = None if isinstance(cb_cal, bool) else int(cb_cal.get_active_id()) # type:ignore[attr-defined]
        dialog.destroy()
        cls._dialog_grid = None # so grid & contents are cleaned up
        return res, dest_cal


    @classmethod
    def _add_row(cls, txt:str, style:str=None, halign:Gtk.Align=None) -> None:
        # Add text to next row of grid in the form "`txt`"
        lab = Gtk.Label(txt)
        if style:
            lab.get_style_context().add_class(style)
        if halign:
            lab.set_halign(halign)
        lab.set_yalign(0.5)
        cls._dialog_grid.attach(lab, 0,cls._dialog_y, 2,1) # spans 2 columns
        cls._dialog_y += 1


    @classmethod
    def _add_row_prop(cls, label:str, cont:str) -> None:
        # Add text to next row of grid in the form "`label` `cont`"
        propnm_lab = Gtk.Label(label)
        propnm_lab.set_halign(Gtk.Align.END)
        propnm_lab.set_yalign(0.5)
        propnm_lab.get_style_context().add_class(GUI.STYLE_TXTLABEL)
        cls._dialog_grid.attach(propnm_lab, 0,cls._dialog_y, 1,1)

        cont_lab = Gtk.Label(cont)
        cont_lab.set_halign(Gtk.Align.START)
        cont_lab.set_yalign(0.5)
        cont_lab.set_xalign(0)
        cont_lab.get_style_context().add_class(GUI.STYLE_TXTPROP)
        cls._dialog_grid.attach(cont_lab, 1,cls._dialog_y, 1,1)

        cls._dialog_y += 1


    @classmethod
    def _add_row_widget(cls, label:str, wid:Gtk.Widget) -> None:
        # Add label + widget to next row of grid
        propnm_lab = Gtk.Label(label)
        propnm_lab.set_halign(Gtk.Align.END)
        propnm_lab.set_yalign(0.5)
        propnm_lab.get_style_context().add_class(GUI.STYLE_TXTLABEL)
        cls._dialog_grid.attach(propnm_lab, 0,cls._dialog_y, 1,1)

        cls._dialog_grid.attach(wid, 1,cls._dialog_y, 1,1)
        cls._dialog_y += 1



# -*- coding: utf-8 -*-
#
# pygenda_gui.py
# Top-level GUI code and shared elements (e.g. menu, soft keys)
#
# Copyright (C) 2022-2025 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.


from gi import require_version as gi_require_version
gi_require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GdkPixbuf, GLib, Gio

from datetime import date as dt_date, time as dt_time, datetime as dt_datetime, timedelta
from icalendar import Calendar as iCalendar, Event as iEvent, Todo as iTodo
from importlib import import_module
from os import path as ospath
from sys import stderr
from pathlib import Path
import signal
import ctypes
from typing import Optional, Tuple, List, Union, Any, Type

# for internationalisation/localisation
import locale
_ = locale.gettext

# pygenda components
from .pygenda_config import Config
from .pygenda_calendar import Calendar
from .pygenda_widgets import WidgetDate
from .pygenda_util import guess_date_ord_from_locale,guess_date_sep_from_locale,guess_time_sep_from_locale, guess_date_fmt_text_from_locale, datetime_to_date
from .pygenda_version import __version__

# Dialog classes - imported in _init_dialogs()
EventDialogController = None # type:Any
TodoDialogController = None # type:Any
FindDialogController = None # type:Any
EntryPropertiesDialog = None # type:Any
EventPropertyBeyondEditDialog = None # type:Any
TodoPropertyBeyondEditDialog = None # type:Any
ImportController = None # type:Any


# Singleton class for top-level GUI control
class GUI:
    _CSS_FILE_APP = '{:s}/css/pygenda.css'.format(ospath.dirname(__file__))
    _CSS_FILE_USER = GLib.get_user_config_dir() + '/pygenda/pygenda.css'
    _LOCALE_DIR = '{:s}/locale/'.format(ospath.dirname(__file__))
    _VIEWS = ('Week','Year','Todo') # Order gives order in menus
    SPINBUTTON_INC_KEY = (Gdk.KEY_plus,Gdk.KEY_greater)
    SPINBUTTON_DEC_KEY = (Gdk.KEY_minus,Gdk.KEY_less)
    STYLE_TXTLABEL = 'tlabel'
    STYLE_TXTPROP = 'plabel'
    STYLE_SECTLABEL = 'sectlabel'
    STYLE_ALERTLABEL = 'alertlabel'
    STYLE_ERR = 'dialog_error'

    views = [] # type: List[Type]
    view_widgets = [] # type: List[Gtk.Widget]
    _view_idx = 0 # take first view as default
    _toggle_view_idx = -1 # used to toggle between views (Esc key)

    _lib_clip = None

    _builder = Gtk.Builder()
    _window = None # type: Gtk.Window
    _is_fullscreen = False
    _box_view_cont = None # type: Gtk.Box
    _eventbox = Gtk.EventBox()

    _menu_elt_fullscreen = None # type: Gtk.Widget
    _menu_elts_entry = None # type: Tuple[Gtk.Widget,...]
    _menu_elts_entry_modify = None # type: Tuple[Gtk.Widget,...]
    _menu_elts_event = None # type: Tuple[Gtk.Widget,...]
    _menu_elts_stat_event = None # type: Tuple[Gtk.Widget,...]
    _menu_elts_stat_todo = None # type: Tuple[Gtk.Widget,...]
    _menu_elts_create = None # type: Tuple[Gtk.Widget,...]

    _image_leave_fs = Gtk.Image.new_from_icon_name('gtk-leave-fullscreen',Gtk.IconSize.MENU)
    _image_enter_fs = None # type: Gtk.Widget

    date_order = ''
    date_formatting_numeric = ''
    date_formatting_text = ''
    date_formatting_text_noyear = ''
    date_formatting_textabb = ''
    date_formatting_textabb_noyear = ''

    _plus_minus_zoom = False

    # Flags to indicate if UI should allow creating events/todos.
    # Set at startup. False if, for example, all calendars are readonly.
    create_events = False
    create_todos = False

    # For startup
    _starting_cal = True
    _loading_indicator = None

    Config.set_defaults('global',{
        'language': '', # use OS language
        'hide_titlebar_when_maximized': False,
        'date_ord': guess_date_ord_from_locale(),
        'date_sep': guess_date_sep_from_locale(),
        'time_sep': guess_time_sep_from_locale(),
        'date_fmt_text': guess_date_fmt_text_from_locale(),
        'date_fmt_text_noyear': '', # construct from date_fmt_text
        'date_fmt_textabb': '', # construct from date_fmt_text
        'date_fmt_textabb_noyear': '', # construct from date_fmtabb_text
        '24hr': False,
        'start_week_day': 0, # 0 = Monday, 6 = Sunday
        'tab_elts_datetime': False,
        'plus_minus_zoom': True,
        })
    Config.set_defaults('startup',{
        'maximize': False,
        'fullscreen': False,
        'view': False,
        })
    Config.set_defaults('softkeys',{
        'display': '',
        'show_icons': True,
        })

    # Constructor
    @classmethod
    def init(cls) -> None:
        # First stage initialisation to bring up the UI.
        # See init_stage2() below for init done after gtk_main loop started.

        # First set the locale, so UI language (e.g. in menu) is correct
        cls._init_locale()

        # Construct GUI from GTK Builder XML UI file
        cls.load_ui_file('main.ui')

        cls._window = cls._builder.get_object('window_main')
        if (not cls._window): # Sanity check
            raise NameError('Main window not found')
        cls._window.set_default_icon_name('x-office-calendar')

        cls._window.set_hide_titlebar_when_maximized(Config.get_bool('global','hide_titlebar_when_maximized'))
        if Config.get_bool('startup','maximize'):
            cls._window.maximize()
        cls._menu_elt_fullscreen = cls._builder.get_object('menuelt-fullscreen')
        cls._image_enter_fs = cls._menu_elt_fullscreen.get_image()
        if Config.get_bool('startup','fullscreen'):
            cls.toggle_fullscreen()

        # Get handles to menu items to enable/disable when on/not on an entry
        cls._menu_elts_entry = cls._get_objs_by_id(('menuelt-copy','menuelt-show-entry-props'))
        cls._menu_elts_entry_modify = cls._get_objs_by_id(('menuelt-cut','menuelt-delete','menuelt-set-status'))
        cls._menu_elts_event = cls._get_objs_by_id(('menuelt-edit-time','menuelt-edit-reps','menuelt-edit-alarm','menuelt-edit-details'))
        cls._menu_elts_stat_event = cls._get_objs_by_id(('menuelt-status-confirmed','menuelt-status-tentative'))
        cls._menu_elts_stat_todo = cls._get_objs_by_id(('menuelt-status-needsaction','menuelt-status-inprocess','menuelt-status-completed'))
        cls._menu_elts_create = cls._get_objs_by_id(('menuelt-paste',))

        # Handle SIGINT (e.g. from ctrl+C) etc.
        GLib.unix_signal_add(GLib.PRIORITY_DEFAULT_IDLE, signal.SIGINT, cls.exit)
        GLib.unix_signal_add(GLib.PRIORITY_DEFAULT_IDLE, signal.SIGTERM, cls.exit)

        # Connect signals now, so clicking on [X] in window exits application
        HANDLERS = {
            'window_main delete': cls.exit,
            'menuitem_import': cls.handler_import,
            'menuitem_quit': cls.exit,
            'menuitem_cut': cls.cut_request,
            'menuitem_copy': cls.copy_request,
            'menuitem_paste': cls.paste_request,
            'menuitem_deleteentry': cls.delete_request,
            'menuitem_newevent': cls.handler_newevent,
            'menuitem_newtodo': cls.handler_newtodo,
            'menuitem_show_entry_props': cls.handler_showentryprops,
            'menuitem_edittime': cls.handler_edittime,
            'menuitem_editrepeats': cls.handler_editrepeats,
            'menuitem_editalarm': cls.handler_editalarm,
            'menuitem_editdetails': cls.handler_editdetails,
            'menuitem_stat_none': lambda a: cls.handler_stat_toggle(None),
            'menuitem_stat_confirmed': lambda a: cls.handler_stat_toggle('CONFIRMED'),
            'menuitem_stat_canceled': lambda a: cls.handler_stat_toggle('CANCELLED'),
            'menuitem_stat_tentative': lambda a: cls.handler_stat_toggle('TENTATIVE'),
            'menuitem_stat_needsaction': lambda a: cls.handler_stat_toggle('NEEDS-ACTION'),
            'menuitem_stat_inprocess': lambda a: cls.handler_stat_toggle('IN-PROCESS'),
            'menuitem_stat_completed': lambda a: cls.handler_stat_toggle('COMPLETED'),
            'menuitem_switchview': cls.switch_view,
            'menuitem_zoomin': lambda a: cls.zoom(+1),
            'menuitem_zoomout': lambda a: cls.zoom(-1),
            'menuitem_fullscreen': cls.toggle_fullscreen,
            'menuitem_goto': cls.dialog_goto,
            'menuitem_find': cls.handler_find,
            'menuitem_about': cls.dialog_about,
            'button0_clicked': cls.handler_newentry,
            'button1_clicked': cls.switch_view,
            'button2_clicked': cls.dialog_goto,
            'button3_clicked': cls.zoom_button,
            }
        cls._builder.connect_signals(HANDLERS)

        cls._box_view_cont = cls._builder.get_object('box_view_cont')
        if (not cls._box_view_cont): # Sanity check
            raise NameError('View container not found')

        # Add a spinner while the view is still loading
        cls._loading_indicator = Gtk.Box()
        cls._loading_indicator.set_name('view_loading')
        spinner = Gtk.Spinner.new()
        cls._loading_indicator.set_center_widget(spinner)
        cls._box_view_cont.pack_start(cls._loading_indicator, True, True, 0)
        spinner.start()
        cls._loading_indicator.show_all()

        # Setup CSS provider(s) now so "loading" notice is styled
        css_prov = Gtk.CssProvider()
        css_prov.load_from_path(cls._CSS_FILE_APP)
        Gtk.StyleContext.add_provider_for_screen(cls._window.get_screen(), css_prov, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
        try:
            css_prov_u = Gtk.CssProvider()
            css_prov_u.load_from_path(cls._CSS_FILE_USER)
            Gtk.StyleContext.add_provider_for_screen(cls._window.get_screen(), css_prov_u, Gtk.STYLE_PROVIDER_PRIORITY_USER)
        except:
            pass

        # Set position/display of soft keys before showing loading indicator
        cfig_skey = Config.get('softkeys','display').lower()
        bbut = cls._builder.get_object('box_buttons')
        if cfig_skey == 'hide':
            bbut.hide()
        elif cfig_skey == 'left':
            cls._box_view_cont.reorder_child(bbut,0)
        else:
            cls._box_view_cont.reorder_child(bbut,1)

        if not Config.get_bool('softkeys','show_icons'):
           bbut.foreach(lambda a: a.get_child().get_children()[0].hide())

        # Delay further initialisation so we can display GUI/window early
        cls._window.show()
        GLib.idle_add(cls.init_stage2)


    @classmethod
    def _get_objs_by_id(cls, id_list:Tuple[str,...]) -> Tuple[Gtk.Widget,...]:
        # Helper function to get a tuple of widgets by id
        return tuple((cls._builder.get_object(id) for id in id_list))


    @classmethod
    def load_ui_file(cls, gfile:str) -> None:
        # Load GUI elements from GTK Builder XML UI file in ui directory
        fullname = ospath.dirname(__file__) + '/ui/' + gfile
        r = cls._builder.add_from_file(fullname)
        if not r:
            print('Error loading UI file '+gfile, file=stderr)
            Gtk.main_quit()


    @staticmethod
    def init_cal(task:Gio.Task, src_obj, t_data, cancel:Gio.Cancellable)->None:
        # Wrapper around Calendar.init(), so we can call it in a GTask thread.
        try:
            Calendar.init()
        except:
            print('Error while creating calendar(s)', file=stderr)
            Gtk.main_quit()
            raise
        task.return_int(0)

    @classmethod
    def init_cal_callback(cls, data, task:Gio.Task) -> None:
        # Callback called to signal that init_cal() has finished.
        cls._starting_cal = False

    @staticmethod
    def init_cal_iosc(job:Gio.IOSchedulerJob, cancel:Gio.Cancellable, u_data) -> bool:
        # Wrapper around Calendar.init, so we can call it in an IOScheduler.
        # Fallback for Gio version < 2.36 (hence needed for Gemini).
        try:
            Calendar.init()
        except:
            print('Error while creating calendar(s)', file=stderr)
            Gtk.main_quit()
            raise
        finally:
            GUI._starting_cal = False
        return False # indicates task complete


    @classmethod
    def init_stage2(cls) -> None:
        # Second-stage initialisation.
        # This is run in the gtk_main thread after basic UI is displayed.

        # First, initialise calendar connector.
        # Do this outside main thread, so UI remains active/responsive.
        ci_cancel = Gio.Cancellable.new()
        if 'run_in_thread' in dir(Gio.Task):
            # Preferred way to run in separate thread.
            task = Gio.Task.new(cancellable=ci_cancel, callback=cls.init_cal_callback)
            task.run_in_thread(cls.init_cal)
        else:
            # Deprecated, but needed to run on Gemini.
            Gio.io_scheduler_push_job(cls.init_cal_iosc, None, GLib.PRIORITY_DEFAULT, ci_cancel)

        # Initialise some things in main thread while calendar loading
        # !! Should probably move these as GIL means they aren't async
        cls._init_clipboard()
        cls._init_date_format()

        # Wait for calendar to finish initialising before doing views
        while cls._starting_cal:
            if Gtk.main_iteration_do(False):
                # True => main_quit() has been called
                ci_cancel.cancel()
                return
            GLib.usleep(8000) # microsecs; low->smooth spin, high->fast start
        del(cls._starting_cal) # Flag no longer needed

        # Get rid of loading indicator, but store position for view placement
        view_pos = cls._box_view_cont.child_get_property(cls._loading_indicator, 'position')
        cls._box_view_cont.remove(cls._loading_indicator)
        del(cls._loading_indicator) # should be only remaining reference

        # Check that at least one calendar initialised/connected properly
        if len(Calendar.calConnectors) == 0:
            print('Error: No calendars created', file=stderr)
            Gtk.main_quit()
            return

        # Check to see if calendar read-only settings means that we can't
        # create any events or to-dos. Disable menu items etc accordingly.
        cls.create_events = len(Calendar.calendar_displaynames_event_rw())!=0
        cls.create_todos = len(Calendar.calendar_displaynames_todo_rw())!=0
        if not cls.create_events:
            cls._builder.get_object('menuelt-new-event').set_sensitive(False)
            cls._builder.get_object('button0').set_sensitive(False)
        if not cls.create_todos:
            cls._builder.get_object('menuelt-new-todo').set_sensitive(False)
        if not cls.create_events and not cls.create_todos:
            cls._builder.get_object('menuelt-paste').set_sensitive(False)
            cls._builder.get_object('menuelt-import').set_sensitive(False)

        cls._init_views()

        # If view set in config, set index
        vw = Config.get('startup','view')
        if vw:
            for ii in range(len(GUI._VIEWS)):
                if GUI._VIEWS[ii].lower() == vw:
                    cls._view_idx = ii
                    break

        # If date set from command line, jump there now
        if Config.date:
            for i in range(cls._view_idx, cls._view_idx+len(cls.views)):
                if cls.views[i%len(cls.views)].cursor_set_date(Config.date):
                    break

        cls._box_view_cont.pack_start(cls._eventbox, True, True, 0)
        cls._box_view_cont.reorder_child(cls._eventbox, view_pos)
        cls._eventbox.add(cls.view_widgets[cls._view_idx])
        cls._eventbox.connect('key-press-event', cls.keypress)
        cls._eventbox.connect('key-release-event', cls.keyrelease)
        cls.view_widgets[cls._view_idx].grab_focus() # so it gets keypresses
        del(cls._box_view_cont) # don't need this anymore

        cls._init_config()
        cls._init_dialogs()

        # Add functionality to spinbuttons not provided by GTK
        cls._init_spinbuttons()
        cls._init_comboboxes()
        cls._init_entryboxes()
        cls._init_switches()

        # Menu bar & softkey bar made insensitive in .ui for startup.
        # We make them sensitive here before activating view.
        cls._builder.get_object('menu_bar').set_sensitive(True)
        cls._builder.get_object('box_buttons').set_sensitive(True)

        cls.view_redraw(True) # Draw active view, including entries
        cls._eventbox.show_all()


    @classmethod
    def _init_config(cls) -> None:
        # Read config settings at startup
        cls._plus_minus_zoom = Config.get_bool('global','plus_minus_zoom')


    @staticmethod
    def _init_dialogs() -> None:
        # Import and initialise dialog classes.
        # Doing imports here avoids circular dependencies.
        global EventDialogController, EventPropertyBeyondEditDialog, TodoDialogController, TodoPropertyBeyondEditDialog, EntryPropertiesDialog, FindDialogController, ImportController

        from .pygenda_dialog_event import EventDialogController, EventPropertyBeyondEditDialog
        EventDialogController.init()

        from .pygenda_dialog_todo import TodoDialogController, TodoPropertyBeyondEditDialog
        TodoDialogController.init() # Need to do this after Todo View init

        from .pygenda_dialog_find import FindDialogController
        FindDialogController.init()

        from .pygenda_dialog_entryprops import EntryPropertiesDialog
        from .pygenda_dialog_import import ImportController


    @classmethod
    def _init_locale(cls) -> None:
        # Initialise language from config or OS settings
        lang = Config.get('global', 'language')
        if not lang:
            lang = locale.getlocale()
            if lang is None or (isinstance(lang, tuple) and lang[0] is None):
                print("Warning: No system language found, falling back to English. This could cause errors later (e.g. opening dialogs) if English is not installed on the system. Perhaps change Pygenda user.ini file to use an installed language.", file=stderr)
                lang = 'en'
        if isinstance(lang,str) and '.' not in lang:
            # Need to include encoding
            lang = (lang,'UTF-8')
        try:
            locale.setlocale(locale.LC_ALL, lang)
        except locale.Error:
            lang_dname = lang if isinstance(lang,str) else '.'.join(lang)
            print("Error loading locale {:s} (not installed on OS?)".format(lang_dname), file=stderr)
            if Path('/etc/debian_version').exists():
                # Add extra info for users on Debian-like systems
                print(" On Debian try: uncomment corresponding line in /etc/locale.gen & run locale-gen", file=stderr)
            print(" Trying to continue with default locale...", file=stderr)
        locale.bindtextdomain('pygenda', cls._LOCALE_DIR)
        locale.textdomain('pygenda')


    @classmethod
    def _init_clipboard(cls) -> None:
        # Load clipboard helper library if available, or set to None
        try:
            libclip_file = '{:s}/libpygenda_clipboard.so'.format(ospath.dirname(__file__))
            cls._lib_clip = ctypes.CDLL(libclip_file)
        except:
            print('Notice: Clipboard library not found, using fallback', file=stderr)


    @classmethod
    def _init_date_format(cls) -> None:
        # Initialise date formatting strings from config
        cls.date_order = cls._date_order_from_config()

        # Make date_formatting_numeric - a format string like '%Y-$m-$d'
        dto_tmp = cls.date_order.replace('M','m').replace('D','d')
        cls.date_formatting_numeric = '%{0:s}{sep:s}%{1:s}{sep:s}%{2:s}'.format(dto_tmp[0],dto_tmp[1],dto_tmp[2],sep=Config.get('global','date_sep'))

        # Make format strings for longer formats, e.g. "Mon Dec 31, 2001"
        cls.date_formatting_text = Config.get('global','date_fmt_text')

        # Other format strings might be constructed if not set in config file
        # date_formatting_text_noyear
        cst = Config.get('global','date_fmt_text_noyear')
        cls.date_formatting_text_noyear = ' '.join(cls.date_formatting_text.replace('%y','%Y').replace('%Y,',' ').replace(', %Y',' ').replace('%Y','').split()) if not cst else cst

        # date_formatting_textabb
        cst = Config.get('global','date_fmt_textabb')
        cls.date_formatting_textabb = cls.date_formatting_text.replace('%A','%a').replace('%B','%b') if not cst else cst

        # date_formatting_textabb_noyear
        cst = Config.get('global','date_fmt_textabb_noyear')
        cls.date_formatting_textabb_noyear = ' '.join(cls.date_formatting_textabb.replace('%y','%Y').replace('%Y,',' ').replace(', %Y',' ').replace('%Y','').split()) if not cst else cst


    @staticmethod
    def _date_order_from_config() -> str:
        # Process date order string from config.
        # Converts, e.g. 'YYYY-MM-DD' -> 'YMD' & checks output is valid.
        raw = Config.get('global','date_ord').upper()
        ret = ''
        for ch in raw:
            if ch in 'YMD' and ch not in ret:
                ret += ch
        assert(len(ret)==3)
        return ret


    @classmethod
    def _init_views(cls) -> None:
        # Get Gtk Widgets for views; add view switching options to menu
        for v in GUI._VIEWS:
            m = import_module('.pygenda_view_{:s}'.format(v.lower()),package='pygenda')
            cls.views.append(getattr(m, 'View_{:s}'.format(v)))
            cls.view_widgets.append(cls.views[-1].init())
            cls.view_widgets[-1].get_style_context().add_class('view')

        # Add views to menu, so the user can switch to them
        menu_views_list = cls._builder.get_object('menu_views_list')
        accel_gp = Gtk.AccelGroup()
        cls._window.add_accel_group(accel_gp)
        for i in range(len(cls.views)):
            m = Gtk.MenuItem(cls.views[i].view_name())
            m.set_use_underline(True)
            akey = cls.views[i].accel_key()
            if akey:
                m.add_accelerator('activate', accel_gp, akey, Gdk.ModifierType.SHIFT_MASK|Gdk.ModifierType.CONTROL_MASK, Gtk.AccelFlags.VISIBLE)
            m.connect('activate',cls.switch_view,i)
            m.show()
            menu_views_list.add(m)


    @classmethod
    def _init_spinbuttons(cls) -> None:
        # Connect spinbutton keypress event to handler for extra features.
        # All spinbuttons are part of the Event dialog, so this could go
        # in the EventDialogController class. However, in theory spinbuttons
        # can be anywhere, so putting this in the general GUI class.
        for sb_id in ('allday_count','repeat_interval','repeat_occurrences'):
            sb = cls._builder.get_object(sb_id)
            sb.connect('key-press-event', cls._spinbutton_keypress)
            sb.connect('focus-out-event', cls._focusout_unhighlight)


    @classmethod
    def _init_comboboxes(cls) -> None:
        # Connect ComboBox events to handlers for extra features.
        for cb_id in ('combo_event_calendar','combo_repeat_type','combo_repeaton_year','combo_repeaton_month','combo_status','combo_todo_list','combo_todo_calendar','combo_todo_priority','combo_todo_status'):
            cb = cls._builder.get_object(cb_id)
            cb.connect('key-press-event', cls._combobox_keypress)


    @classmethod
    def _init_entryboxes(cls) -> None:
        # Connect Entry textbox events to handlers for extra features.
        for eb_id in ('entry_dialogevent_desc','entry_dialogevent_location','entry_dialogtodo_desc'):
            eb = cls._builder.get_object(eb_id)
            eb.connect('focus-out-event', cls._focusout_unhighlight)


    @classmethod
    def _init_switches(cls) -> None:
        # Connect Switch events to handlers for extra features.
        for sw_id in ('dialogevent_switch_alarm','switch_todo_duedate'):
            sw = cls._builder.get_object(sw_id)
            sw.connect('key-press-event', cls._switch_keypress)


    @classmethod
    def set_menu_elts(cls, on_event:bool=False, on_todo:bool=False, read_only:bool=False, can_create_here=False) -> None:
        # Called from Views as the cursor is moved. Enables/disables/hides
        # menu items appropriate for the current cursor item.
        for id in cls._menu_elts_entry:
            id.set_sensitive(on_event or on_todo)
        for id in cls._menu_elts_entry_modify:
            id.set_sensitive((not read_only) and (on_event or on_todo))
        for id in cls._menu_elts_event:
            id.set_sensitive((not read_only) and on_event)
        for id in cls._menu_elts_create:
            id.set_sensitive(can_create_here)
        if on_event:
            for id in cls._menu_elts_stat_event:
                id.show()
            for id in cls._menu_elts_stat_todo:
                id.hide()
        elif on_todo:
            for id in cls._menu_elts_stat_event:
                id.hide()
            for id in cls._menu_elts_stat_todo:
                id.show()


    @classmethod
    def keypress(cls, wid:Gtk.Widget, ev:Gdk.EventKey) -> bool:
        # Called whenever a key is pressed/repeated when View in focus
        if cls._plus_minus_zoom:
            if ev.keyval==Gdk.KEY_plus:
                cls.zoom(+1)
                return True # handled
            elif ev.keyval==Gdk.KEY_minus:
                cls.zoom(-1)
                return True # handled
        if ev.keyval==Gdk.KEY_Escape:
            if cls._toggle_view_idx >= 0:
                cls.switch_view(None, cls._toggle_view_idx)
        else:
            cls.views[cls._view_idx].keypress(wid,ev)
        return True # event handled - don't propagate


    @classmethod
    def keyrelease(cls, wid:Gtk.Widget, ev:Gdk.EventKey) -> bool:
        # Called whenever a key is released
        cls.views[cls._view_idx].keyrelease(wid,ev)
        return True # event handled - don't propagate


    @staticmethod
    def _spinbutton_keypress(wid:Gtk.SpinButton, ev:Gdk.EventKey) -> bool:
        # Called to handle extra spinbutton keyboard controls
        shiftdown = ev.state&Gdk.ModifierType.SHIFT_MASK
        if ev.keyval in GUI.SPINBUTTON_INC_KEY or (shiftdown and ev.keyval==Gdk.KEY_Up):
            wid.update() # So if user types "5" then "+" value changes to "6"
            wid.spin(Gtk.SpinType.STEP_FORWARD, 1)
            return True # done, don't propagate
        if ev.keyval in GUI.SPINBUTTON_DEC_KEY or (shiftdown and ev.keyval==Gdk.KEY_Down):
            wid.update()
            wid.spin(Gtk.SpinType.STEP_BACKWARD, 1)
            return True # done
        if ev.keyval==Gdk.KEY_Up:
            return wid.get_toplevel().child_focus(Gtk.DirectionType.UP)
        if ev.keyval==Gdk.KEY_Down:
            return wid.get_toplevel().child_focus(Gtk.DirectionType.DOWN)
        return False # unhandled, so propagate event


    @staticmethod
    def _combobox_keypress(wid:Gtk.ComboBox, ev:Gdk.EventKey, resp:Gtk.ResponseType=Gtk.ResponseType.OK) -> bool:
        # Called to handle extra combobox keyboard controls
        # BUG!! This is not called when the combobox is in "popout" state
        if ev.keyval==Gdk.KEY_Return:
            # Manually trigger default event on dialog box
            dlg = wid.get_toplevel()
            if dlg:
                dlg.response(resp)
            return True # done

        mdl = wid.get_model()
        count = mdl.iter_n_children()
        a = wid.get_active()
        shiftdown = ev.state&Gdk.ModifierType.SHIFT_MASK
        if ev.keyval in GUI.SPINBUTTON_INC_KEY or (shiftdown and ev.keyval==Gdk.KEY_Down):
            a = (a+1)%count
            wid.set_active(a)
            return True # done
        if ev.keyval in GUI.SPINBUTTON_DEC_KEY or (shiftdown and ev.keyval==Gdk.KEY_Up):
            if a<0:
                a = 0
            a = (a-1)%count
            wid.set_active(a)
            return True # done, don't propagate

        if ev.keyval==Gdk.KEY_Up:
            return wid.get_toplevel().child_focus(Gtk.DirectionType.UP)
        if ev.keyval==Gdk.KEY_Down:
            return wid.get_toplevel().child_focus(Gtk.DirectionType.DOWN)
        if ev.keyval in (Gdk.KEY_Left,Gdk.KEY_Right):
            # Return - otherwise detected as alphabetic (not sure why!)
            return False

        ch = chr(ev.keyval)
        if ch.isalpha() or ch.isdigit():
            ch = ch.lower()
            # Search for ch in first characters of combobox values
            it = mdl.iter_nth_child(None,a)
            for i in range(1,count):
                it = mdl.iter_next(it)
                if it is None: # at end of list, loop to top!
                    it = mdl.get_iter_first()
                ch1 = mdl.get(it,0)[0][0].lower()
                if ch==ch1:
                    wid.set_active((a+i)%count)
                    break
            return True # done, don't propagate

        return False # propagate event


    @staticmethod
    def _switch_keypress(wid:Gtk.Switch, ev:Gdk.EventKey) -> bool:
        # Called to handle extra switch keyboard controls
        if ev.keyval==Gdk.KEY_Left:
            if wid.get_active():
                wid.activate()
                return True
        elif ev.keyval==Gdk.KEY_Right:
            if not wid.get_active():
                wid.activate()
                return True
        return False # unhandled, so propagate event


    @staticmethod
    def _focusout_unhighlight(wid:Gtk.Widget, ev:Gdk.EventKey) -> bool:
        # Called to handle remove highlight from entry box/spinbutton
        # when focus moves to another widget
        dlg = wid.get_toplevel()
        if dlg.get_focus() != wid:
            wid.select_region(0,0) # remove highlight
        return False # propagate event


    @classmethod
    def view_redraw(cls, en_changes:bool=False) -> None:
        # Redraw the currently active view.
        # en_changes: bool, True if displayed entries need updating too
        cls.views[cls._view_idx].redraw(en_changes=en_changes)


    @classmethod
    def switch_view(cls, wid:Gtk.Widget, idx:int=None, redraw:bool=True) -> bool:
        # Callback from UI widget (e.g. menu, softkey) to change view.
        # idx = index of new view (otherwise goes to next view in list)
        if idx is None:
            # Go to next view in list
            cls._toggle_view_idx = cls._view_idx
            cls._view_idx = (cls._view_idx+1)%len(cls.views)
        elif cls._view_idx == idx:
            return True # No change, so skip redraw & don't propagate event
        else:
            cls._toggle_view_idx = cls._view_idx
            cls._view_idx = idx
        cls._eventbox.remove(cls._eventbox.get_child())
        new_view = cls.views[cls._view_idx]
        new_view.renew_display()
        if redraw:
            new_view.redraw(en_changes=True)
        new_wid = cls.view_widgets[cls._view_idx]
        cls._eventbox.add(new_wid)
        new_wid.grab_focus()
        new_wid.show_all()
        return True # don't propagate event


    @classmethod
    def cursor_goto_date(cls, dt:dt_date) -> None:
        # Call view to set current cursor date.
        # If current view does not support setting date (e.g. Todo View)
        # try next view etc., and make successful view active.
        # (N.B.: Does not call redraw - caller needs to handle that.)
        for i in range(cls._view_idx, cls._view_idx+len(cls.views)):
            v = i%len(cls.views)
            if cls.views[v].cursor_set_date(dt): # True if view can show date
                if v != cls._view_idx:
                    cls.switch_view(None, v, redraw=False)
                break


    @classmethod
    def cursor_goto_event(cls, ev:iEvent) -> None:
        # Call view to move cursor to given event.
        # If current view does not support showing event (e.g. Todo View),
        # try next view etc., and make successful view active.
        # (N.B.: Does not call redraw - caller needs to handle that.)
        for i in range(cls._view_idx, cls._view_idx+len(cls.views)):
            v = i%len(cls.views)
            if cls.views[v].cursor_goto_event(ev): # True if view can show event
                if v != cls._view_idx:
                    cls.switch_view(None, v, redraw=False)
                break


    @classmethod
    def cursor_goto_todo(cls, todo:iTodo, list_idx:int) -> None:
        # Call view to move cursor to given todo in given todo list.
        # If current view does not support displaying todo, try
        # the next view etc., and make successful view active.
        # (N.B.: Does not call redraw - caller needs to handle that.)
        for i in range(cls._view_idx, cls._view_idx+len(cls.views)):
            v = i%len(cls.views)
            if cls.views[v].cursor_goto_todo(todo, list_idx): # True if view can show todo
                if v != cls._view_idx:
                    cls.switch_view(None, v, redraw=False)
                break


    # Main
    @classmethod
    def main(cls) -> None:
        # Run the main Gtk loop
        Gtk.main()


    # Signal handling functions
    @classmethod
    def exit(cls, *args) -> bool:
        # Callback for various types of exit signal (command line, menus...)
        Gtk.main_quit()
        return True # don't propagate event

    @classmethod
    def handler_newevent(cls, *args) -> bool:
        # Callback for "Create new event" signal (menu)
        date = cls.views[cls._view_idx].cursor_date()
        EventDialogController.new_event(date=date)
        return True # don't propagate event

    @classmethod
    def handler_newtodo(cls, *args) -> bool:
        # Callback for "Create new todo" signal (menu)
        lst = cls.views[cls._view_idx].cursor_todo_list()
        TodoDialogController.new_todo(list_idx=lst)
        return True # don't propagate event

    @classmethod
    def handler_newentry(cls, *args) -> bool:
        # Callback for "Create new entry" signal (softkey)
        # Create new event or new todo, depending on the View.
        if cls.views[cls._view_idx].default_entry_is_todo():
            return cls.handler_newtodo(cls, *args)
        return cls.handler_newevent(cls, *args)

    @classmethod
    def handler_import(cls, *args) -> bool:
        # Handler to implement "Import file" from GUI, e.g. clicked in menu
        ImportController.import_flow()
        return True # don't propagate event

    @classmethod
    def handler_find(cls, *args) -> bool:
        # Callback for find signal (menu)
        FindDialogController.find()
        return True # don't propagate event

    @classmethod
    def handler_showentryprops(cls, *args) -> bool:
        # Callback for "show entry properties" signal (menu item)
        en = cls.views[cls._view_idx].get_cursor_entry()
        if en:
            cls.dialog_showentryprops(en)
        return True # don't propagate event

    @classmethod
    def handler_edittime(cls, *args) -> bool:
        # Callback for "edit entry time" signal (menu item)
        en = cls.views[cls._view_idx].get_cursor_entry()
        if en:
            cls.edit_or_display_event(en, EventDialogController.TAB_TIME)
        return True # don't propagate event

    @classmethod
    def handler_editrepeats(cls, *args) -> bool:
        # Callback for "edit entry repeats" signal (menu item)
        en = cls.views[cls._view_idx].get_cursor_entry()
        if en:
            cls.edit_or_display_event(en, EventDialogController.TAB_REPEATS)
        return True # don't propagate event

    @classmethod
    def handler_editalarm(cls, *args) -> bool:
        # Callback for "edit entry alarms" signal (menu item)
        en = cls.views[cls._view_idx].get_cursor_entry()
        if en:
            cls.edit_or_display_event(en, EventDialogController.TAB_ALARM)
        return True # don't propagate event

    @classmethod
    def handler_editdetails(cls, *args) -> bool:
        # Callback for "edit entry details" signal (menu item)
        en = cls.views[cls._view_idx].get_cursor_entry()
        if en:
            cls.edit_or_display_event(en, EventDialogController.TAB_DETAILS)
        return True # don't propagate event


    @classmethod
    def edit_or_display_event(cls, en:iEvent, subtab:int=None) -> None:
        # Bring up dialog to edit event, or show details if not editable
        if not Calendar.calendar_readonly(en):
            try:
                EventDialogController.edit_event(en, subtab)
            except EventPropertyBeyondEditDialog as exc:
                print('Warning: {:s} - showing entry properties'.format(str(exc)), file=stderr)
                cls.dialog_showentryprops(en)
        else:
            cls.dialog_showentryprops(en)


    @classmethod
    def edit_or_display_todo(cls, todo:iTodo, list_idx:int=None) -> None:
        # Bring up dialog to edit todo item, or show details if not editable
        if not Calendar.calendar_readonly(todo):
            try:
                TodoDialogController.edit_todo(todo, list_idx)
            except TodoPropertyBeyondEditDialog as exc:
                print('Warning: {:s} - showing todo properties'.format(str(exc)), file=stderr)
                cls.dialog_showentryprops(todo)
        else:
            cls.dialog_showentryprops(todo)


    @classmethod
    def handler_stat_toggle(cls, stat:Optional[str]) -> bool:
        # Handle signals from menu to change current entry's status.
        # Paramenter stat is None or text string for status (eg 'CONFIRMED').
        en = cls.views[cls._view_idx].get_cursor_entry()
        if en:
            Calendar.set_toggle_status_entry(en, stat)
            cls.view_redraw(en_changes=True)
        return True # don't propagate event


    @classmethod
    def delete_request(cls, *args) -> bool:
        # Callback to implement "delete" from GUI, e.g. backspace key pressed
        en = cls.views[cls._view_idx].get_cursor_entry()
        if en is not None:
            cls.dialog_deleteentry(en)
        return True # don't propagate event


    @classmethod
    def cut_request(cls, *args) -> bool:
        # Handler to implement "cut" from GUI, e.g. cut clicked in menu
        en = cls.views[cls._view_idx].get_cursor_entry()
        if en and 'SUMMARY' in en:
            if 'RRULE' in en: # repeating entry
                # Need to think about how to implement this from UI side.
                # Problem: Does user expect single occurrence to be cut, or all?
                # Maybe bring up dialog "Cut single occurrence, or all repeats?"
                # Then, do we do the came for Copying repeating entries?
                # How do we adapt repeats when moved to a different date?
                print('Warning: Cutting repeating entries not implemented', file=stderr)
                return True
            cls._put_entry_on_clipboard(en)
            Calendar.delete_entry(en)
            cls.view_redraw(en_changes=True)
        return True # don't propagate event


    @classmethod
    def copy_request(cls, *args) -> bool:
        # Handler to implement "copy" from GUI, e.g. copy clicked in menu
        en = cls.views[cls._view_idx].get_cursor_entry()
        if en and 'SUMMARY' in en:
            cls._put_entry_on_clipboard(en)
        return True # don't propagate event


    @classmethod
    def _put_entry_on_clipboard(cls, en:Union[iEvent,iTodo]) -> None:
        calbuf = en.to_ical()
        if cls._lib_clip is None:
            cb = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
            cb.set_text(calbuf.decode('utf-8'), -1)
        else:
            txtbuf = bytes(en['SUMMARY'], 'utf-8')
            cls._lib_clip.set_cb(ctypes.create_string_buffer(txtbuf),ctypes.create_string_buffer(calbuf))


    @classmethod
    def paste_request(cls, *args) -> bool:
        # Handler to implement "paste" from GUI, e.g. paste clicked in menu
        cb = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)

        # First, we try requesting a 'text/calendar' type from the clipboard
        sdat = cb.wait_for_contents(Gdk.Atom.intern('text/calendar', False))
        try:
            ical = iCalendar.from_ical(sdat.get_data())
            en = ical.walk()[0]
            cls.views[cls._view_idx].paste_entry(en)
            cls.view_redraw(en_changes=True)
        except:
            # Fallback: request plain text from clipboard
            txt = cb.wait_for_text()
            if txt: # might be None
                if cls._lib_clip is None:
                    # "Text" might be ical data, let's check...
                    try:
                        ical = iCalendar.from_ical(txt)
                        en = ical.walk()[0]
                        cls.views[cls._view_idx].paste_entry(en)
                        cls.view_redraw(en_changes=True)
                        return True
                    except:
                        pass
                # We got text. Either clipboard library is present, or
                # there was an exception trying to treat it as ical data.
                # So, let's just treat it as plain text...
                txt = cls._sanitise_pasted_text(txt)
                # Type of entry created depends on View, so call View paste fn
                if txt:
                    cls.views[cls._view_idx].paste_text(txt)
        return True # don't propagate event


    @staticmethod
    def _sanitise_pasted_text(txt:str) -> str:
        # Clean up text from clipboard so it's suitable as an entry summary
        txt = txt.strip()
        txt = txt.replace('\n',' ')
        txt = txt.replace('\t',' ')
        return txt


    @classmethod
    def dialog_deleteentry(cls, en:iEvent) -> None:
        # Dialog to implement "delete" from GUI, e.g. backspace key
        dialog = Gtk.Dialog(title=_('Delete Entry'), parent=cls._window,
            flags=Gtk.DialogFlags.MODAL|Gtk.DialogFlags.DESTROY_WITH_PARENT,
            buttons=(Gtk.STOCK_CANCEL, Gtk.ResponseType.CLOSE, Gtk.STOCK_DELETE, Gtk.ResponseType.APPLY))
        if 'RRULE' in en:
            # repeating entry - clarify what is being deleted
            # !! We should really ask if user wants to delete all/single etc.
            l_template = _(u'Delete all repeats:\n“{:s}”?')
        else:
            l_template = _(u'Delete entry:\n“{:s}”?')
        lab = Gtk.Label(l_template.format(en['SUMMARY'] if 'SUMMARY' in en else u' ')) # narrow space
        if (not dialog or not lab): # Sanity check
            raise NameError('Dialog Delete creation failure')
        dialog.set_resizable(False)
        lab.set_justify(Gtk.Justification.CENTER)
        dialog.get_content_area().add(lab)
        dialog.set_default_response(Gtk.ResponseType.APPLY)#Enter action
        dialog.show_all()
        response = dialog.run()
        dialog.destroy()
        if response == Gtk.ResponseType.APPLY:
            Calendar.delete_entry(en)
            cls.view_redraw(en_changes=True)


    @classmethod
    def dialog_showentryprops(cls, entry:Union[iEvent,iTodo]) -> None:
        # Shows dialog with entry details.
        # Used if requested from menu or, e.g., for read-only entries.
        dialog = EntryPropertiesDialog(entry, parent=cls._window)
        response = dialog.run()
        dialog.destroy()


    @classmethod
    def dialog_goto(cls, *args) -> bool:
        # Called to implement "go to" from GUI, e.g. button
        dialog = Gtk.Dialog(title=_('Go To'), parent=cls._window,
            flags=Gtk.DialogFlags.MODAL|Gtk.DialogFlags.DESTROY_WITH_PARENT,
            buttons=(Gtk.STOCK_CANCEL, Gtk.ResponseType.CLOSE))
        wdate = WidgetDate(cls.views[cls._view_idx].cursor_date())
        if (not dialog or not wdate): # Allocation check
            raise NameError('Dialog Goto creation failure')
        wdate.connect('changed', GUI.check_date_fixed)
        dialog.set_resizable(False)
        dialog.get_content_area().add(wdate)
        wdate.set_halign(Gtk.Align.CENTER)
        icon = Gtk.Image()
        icon.set_from_stock(Gtk.STOCK_JUMP_TO, Gtk.IconSize.BUTTON)
        go_but = Gtk.Button(label=_('Go'), image=icon)
        go_but.set_can_default(True)
        dialog.add_action_widget(go_but, Gtk.ResponseType.APPLY)
        go_but.grab_default() # So "Enter" activates go_but
        dialog.show_all()
        while True:
            response = dialog.run()
            dt = wdate.get_approx_date_or_none() # So Feb 30th -> end of Feb
            if response!=Gtk.ResponseType.APPLY or dt is not None:
                break
            # Date is invalid, add error styling
            wdate.get_style_context().add_class(GUI.STYLE_ERR)
        if response == Gtk.ResponseType.APPLY:
            cls.cursor_goto_date(dt)
            cls.view_redraw(en_changes=False)
        dialog.destroy()
        return True # don't propagate event


    @staticmethod
    def check_date_fixed(wid:WidgetDate) -> bool:
        # Removes error highlight if date is valid (e.g. not 30th Feb)
        # Would be nice to make this a method of WidgetDate class.
        # Can be used as a callback, e.g. attached to 'changed' signal
        if wid.get_date_or_none() is not None:
            wid.get_style_context().remove_class(GUI.STYLE_ERR)
        return False # propagate event


    @classmethod
    def dialog_about(cls, *args) -> bool:
        # Display the "About Pygenda" dialog
        dialog = Gtk.AboutDialog(parent=cls._window)
        dialog.set_program_name('Pygenda')
        dialog.set_copyright(u'Copyright © 2022–2025 Matthew Lewis')
        dialog.set_license_type(Gtk.License.GPL_3_0_ONLY)
        github_url = 'https://github.com/semiprime/pygenda'
        dialog.set_website(github_url)
        dialog.set_website_label(_('Source code & documentation: ')+github_url)
        logo = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, True, 8, 1, 1)
        dialog.set_logo(logo)
        dialog.set_authors(('Matthew Lewis',))
        dialog.set_version('version {:s}'.format(__version__))
        dialog.set_comments(_(u'A calendar/agenda application written in Python/GTK3. The UI is inspired by the Agenda apps on the Psion Series 3 and Series 5 PDAs.\nWARNING: This is in-development code, released for testing and feedback. There will be bugs; please report them to: pygenda@semiprime.com.'))
        dialog.add_credit_section(_('Thanks for testing &amp; feedback'),('Christoph (@chaos.social@inlovewithpda) https://chaos.social/@inlovewithpda', 'Edward Hasbrouck', 'Neil Sands'))
        dialog.set_translator_credits(_('translator-credits'))
        dialog.show_all()
        dialog.run()
        dialog.destroy()
        return True # don't propagate event


    @classmethod
    def toggle_fullscreen(cls, *args) -> bool:
        # Callback to toggle fullscreen mode on/off (e.g. from menu)
        cls._is_fullscreen = not cls._is_fullscreen
        if cls._is_fullscreen:
            cls._window.fullscreen()
            cls._menu_elt_fullscreen.set_label('gtk-leave-fullscreen')
            cls._menu_elt_fullscreen.set_image(cls._image_leave_fs)
        else:
            cls._window.unfullscreen()
            cls._menu_elt_fullscreen.set_label('gtk-fullscreen')
            cls._menu_elt_fullscreen.set_image(cls._image_enter_fs)
        return True # don't propagate event


    @classmethod
    def zoom_button(cls, wid:Gtk.Widget, ev:Gdk.EventKey) -> bool:
        # Zoom handler for Zoom softkey
        cls.zoom(-1 if ev.state&(Gdk.ModifierType.SHIFT_MASK|Gdk.ModifierType.CONTROL_MASK) else +1)
        return True # don't propagate event


    @classmethod
    def zoom(cls, inc:int) -> bool:
        # Callback for zoom-in/out menu items
        # Zoom by inc, so zoom-in if inc==+1; zoom-out if inc==-1
        # Call current view to do the work
        cls.views[cls._view_idx].zoom(inc)
        return True # don't propagate event


    @classmethod
    def todo_titles_default_cats(cls) -> Tuple[list,list]:
        # Return titles and default categories of todo lists
        try:
            todo_idx = cls._VIEWS.index('Todo')
        except ValueError:
            return [], []
        tdv = cls.views[todo_idx]
        return tdv._list_titles, tdv._list_default_cats


    @classmethod
    def cursor_date(cls) -> Optional[dt_date]:
        # Returns date (maybe datetime in the future) with cursor.
        # None means no date (e.g. Todo View)
        return cls.views[cls._view_idx].cursor_date() # type:ignore[no-any-return]
